"""Admin routes for presentations."""

import json
from typing import Mapping, Sequence, Tuple

from aiohttp import web
from aiohttp_apispec import (
    docs,
    match_info_schema,
    querystring_schema,
    request_schema,
    response_schema,
)
from marshmallow import ValidationError, fields, validate, validates_schema
from marshmallow.validate import Range

from ....admin.decorators.auth import tenant_authentication
from ....admin.request_context import AdminRequestContext
from ....anoncreds.holder import AnonCredsHolder, AnonCredsHolderError
from ....anoncreds.models.presentation_request import AnonCredsPresentationRequestSchema
from ....anoncreds.models.proof import AnonCredsPresSpecSchema
from ....connections.models.conn_record import ConnRecord
from ....indy.holder import IndyHolder, IndyHolderError
from ....indy.models.cred_precis import IndyCredPrecisSchema
from ....indy.models.proof import IndyPresSpecSchema
from ....indy.models.proof_request import IndyProofRequestSchema
from ....indy.util import generate_pr_nonce
from ....ledger.error import LedgerError
from ....messaging.decorators.attach_decorator import AttachDecorator
from ....messaging.models.base import BaseModelError
from ....messaging.models.openapi import OpenAPISchema
from ....messaging.models.paginated_query import (
    PaginatedQuerySchema,
    get_paginated_query_params,
)
from ....messaging.valid import (
    INDY_EXTRA_WQL_EXAMPLE,
    INDY_EXTRA_WQL_VALIDATE,
    NUM_STR_NATURAL_EXAMPLE,
    NUM_STR_NATURAL_VALIDATE,
    NUM_STR_WHOLE_EXAMPLE,
    NUM_STR_WHOLE_VALIDATE,
    UUID4_EXAMPLE,
    UUID4_VALIDATE,
)
from ....storage.base import DEFAULT_PAGE_SIZE, MAXIMUM_PAGE_SIZE, BaseStorage
from ....storage.error import StorageError, StorageNotFoundError
from ....storage.vc_holder.base import VCHolder
from ....storage.vc_holder.vc_record import VCRecord
from ....utils.tracing import AdminAPIMessageTracingSchema, get_timer, trace_event
from ....vc.ld_proofs import (
    BbsBlsSignature2020,
    EcdsaSecp256r1Signature2019,
    Ed25519Signature2018,
    Ed25519Signature2020,
)
from ....wallet.error import WalletNotFoundError
from ..dif.pres_exch import ClaimFormat, InputDescriptors, SchemaInputDescriptor
from ..dif.pres_proposal_schema import DIFProofProposalSchema
from ..dif.pres_request_schema import DIFPresSpecSchema, DIFProofRequestSchema
from . import problem_report_for_record, report_problem
from .formats.handler import V20PresFormatHandlerError
from .manager import V20PresManager
from .message_types import ATTACHMENT_FORMAT, PRES_20_PROPOSAL, PRES_20_REQUEST, SPEC_URI
from .messages.pres_format import V20PresFormat
from .messages.pres_problem_report import ProblemReportReason
from .messages.pres_proposal import V20PresProposal
from .messages.pres_request import V20PresRequest
from .models.pres_exchange import V20PresExRecord, V20PresExRecordSchema


class V20PresentProofModuleResponseSchema(OpenAPISchema):
    """Response schema for Present Proof Module."""


class V20PresExRecordListQueryStringSchema(PaginatedQuerySchema):
    """Parameters and validators for presentation exchange list query."""

    connection_id = fields.Str(
        required=False,
        metadata={"description": "Connection identifier", "example": UUID4_EXAMPLE},
    )
    thread_id = fields.Str(
        required=False,
        metadata={"description": "Thread identifier", "example": UUID4_EXAMPLE},
    )
    role = fields.Str(
        required=False,
        validate=validate.OneOf(
            [
                getattr(V20PresExRecord, m)
                for m in vars(V20PresExRecord)
                if m.startswith("ROLE_")
            ]
        ),
        metadata={"description": "Role assigned in presentation exchange"},
    )
    state = fields.Str(
        required=False,
        validate=validate.OneOf(
            [
                getattr(V20PresExRecord, m)
                for m in vars(V20PresExRecord)
                if m.startswith("STATE_")
            ]
        ),
        metadata={"description": "Presentation exchange state"},
    )


class V20PresExRecordListSchema(OpenAPISchema):
    """Result schema for a presentation exchange query."""

    results = fields.List(
        fields.Nested(V20PresExRecordSchema()),
        metadata={"description": "Presentation exchange records"},
    )


class V20PresProposalByFormatSchema(OpenAPISchema):
    """Schema for presentation proposal per format."""

    anoncreds = fields.Nested(
        AnonCredsPresentationRequestSchema,
        required=False,
        metadata={"description": "Presentation proposal for anoncreds"},
    )
    indy = fields.Nested(
        IndyProofRequestSchema,
        required=False,
        metadata={"description": "Presentation proposal for indy"},
    )
    dif = fields.Nested(
        DIFProofProposalSchema,
        required=False,
        metadata={"description": "Presentation proposal for DIF"},
    )

    @validates_schema
    def validate_fields(self, data, **kwargs):
        """Validate schema fields: data must have at least one format.

        Args:
            data: The data to validate
            kwargs: Additional keyword arguments

        Raises:
            ValidationError: if data has no formats

        """
        if not any(f.api in data for f in V20PresFormat.Format):
            raise ValidationError(
                "V20PresProposalByFormatSchema requires indy, dif, or both"
            )


class V20PresProposalRequestSchema(AdminAPIMessageTracingSchema):
    """Request schema for sending a presentation proposal admin message."""

    connection_id = fields.Str(
        required=True,
        metadata={"description": "Connection identifier", "example": UUID4_EXAMPLE},
    )
    comment = fields.Str(
        required=False,
        allow_none=True,
        metadata={"description": "Human-readable comment"},
    )
    presentation_proposal = fields.Nested(V20PresProposalByFormatSchema(), required=True)
    auto_present = fields.Boolean(
        required=False,
        dump_default=False,
        metadata={
            "description": (
                "Whether to respond automatically to presentation requests, building"
                " and presenting requested proof"
            )
        },
    )
    auto_remove = fields.Bool(
        required=False,
        dump_default=False,
        metadata={
            "description": (
                "Whether to remove the presentation exchange record on completion"
                " (overrides --preserve-exchange-records configuration setting)"
            )
        },
    )
    trace = fields.Bool(
        required=False,
        metadata={
            "description": "Whether to trace event (default false)",
            "example": False,
        },
    )


class V20PresRequestByFormatSchema(OpenAPISchema):
    """Presentation request per format."""

    anoncreds = fields.Nested(
        AnonCredsPresentationRequestSchema,
        required=False,
        metadata={"description": "Presentation proposal for anoncreds"},
    )
    indy = fields.Nested(
        IndyProofRequestSchema,
        required=False,
        metadata={"description": "Presentation request for indy"},
    )
    dif = fields.Nested(
        DIFProofRequestSchema,
        required=False,
        metadata={"description": "Presentation request for DIF"},
    )

    @validates_schema
    def validate_fields(self, data, **kwargs):
        """Validate schema fields: data must have at least one format.

        Args:
            data: The data to validate
            kwargs: Additional keyword arguments

        Raises:
            ValidationError: if data has no formats

        """
        if not any(f.api in data for f in V20PresFormat.Format):
            raise ValidationError(
                "V20PresRequestByFormatSchema requires at least one of: "
                "anoncreds, indy, dif"
            )


class V20PresCreateRequestRequestSchema(AdminAPIMessageTracingSchema):
    """Request schema for creating a proof request free of any connection."""

    presentation_request = fields.Nested(V20PresRequestByFormatSchema(), required=True)
    comment = fields.Str(required=False, allow_none=True)
    auto_verify = fields.Bool(
        required=False,
        metadata={
            "description": "Verifier choice to auto-verify proof presentation",
            "example": False,
        },
    )
    auto_remove = fields.Bool(
        required=False,
        dump_default=False,
        metadata={
            "description": (
                "Whether to remove the presentation exchange record on completion"
                " (overrides --preserve-exchange-records configuration setting)"
            )
        },
    )
    trace = fields.Bool(
        required=False,
        metadata={
            "description": "Whether to trace event (default false)",
            "example": False,
        },
    )


class V20PresSendRequestRequestSchema(V20PresCreateRequestRequestSchema):
    """Request schema for sending a proof request on a connection."""

    connection_id = fields.Str(
        required=True,
        metadata={"description": "Connection identifier", "example": UUID4_EXAMPLE},
    )


class V20PresentationSendRequestToProposalSchema(AdminAPIMessageTracingSchema):
    """Request schema for sending a proof request bound to a proposal."""

    auto_verify = fields.Bool(
        required=False,
        metadata={
            "description": "Verifier choice to auto-verify proof presentation",
            "example": False,
        },
    )
    auto_remove = fields.Bool(
        required=False,
        dump_default=False,
        metadata={
            "description": (
                "Whether to remove the presentation exchange record on completion"
                " (overrides --preserve-exchange-records configuration setting)"
            )
        },
    )
    trace = fields.Bool(
        required=False,
        metadata={
            "description": "Whether to trace event (default false)",
            "example": False,
        },
    )


class V20PresSpecByFormatRequestSchema(AdminAPIMessageTracingSchema):
    """Presentation specification schema by format, for send-presentation request."""

    anoncreds = fields.Nested(
        AnonCredsPresSpecSchema,
        required=False,
        metadata={"description": "Presentation specification for anoncreds"},
    )
    indy = fields.Nested(
        IndyPresSpecSchema,
        required=False,
        metadata={"description": "Presentation specification for indy"},
    )
    dif = fields.Nested(
        DIFPresSpecSchema,
        required=False,
        metadata={
            "description": (
                "Optional Presentation specification for DIF, overrides the"
                " PresentationExchange record's PresRequest"
            )
        },
    )
    auto_remove = fields.Bool(
        required=False,
        dump_default=False,
        metadata={
            "description": (
                "Whether to remove the presentation exchange record on completion"
                " (overrides --preserve-exchange-records configuration setting)"
            )
        },
    )

    @validates_schema
    def validate_fields(self, data, **kwargs):
        """Validate schema fields: specify exactly one format.

        Args:
            data: The data to validate
            kwargs: Additional keyword arguments

        Raises:
            ValidationError: if data does not have exactly one format.

        """
        if len(data.keys() & {f.api for f in V20PresFormat.Format}) < 1:
            raise ValidationError(
                "V20PresSpecByFormatRequestSchema must specify "
                "at least one presentation format"
            )


class V20CredentialsFetchQueryStringSchema(OpenAPISchema):
    """Parameters and validators for credentials fetch request query string."""

    referent = fields.Str(
        required=False,
        metadata={
            "description": "Proof request referents of interest, comma-separated",
            "example": "1_name_uuid,2_score_uuid",
        },
    )
    start = fields.Str(
        required=False,
        load_default="0",
        validate=NUM_STR_WHOLE_VALIDATE,
        metadata={
            "description": "Start index (DEPRECATED - use offset instead)",
            "strict": True,
            "example": NUM_STR_WHOLE_EXAMPLE,
            "deprecated": True,
        },
    )
    count = fields.Str(
        required=False,
        load_default="10",
        validate=NUM_STR_NATURAL_VALIDATE,
        metadata={
            "description": "Maximum number to retrieve (DEPRECATED - use limit instead)",
            "example": NUM_STR_NATURAL_EXAMPLE,
            "deprecated": True,
        },
    )
    limit = fields.Int(
        required=False,
        validate=Range(min=1, max=MAXIMUM_PAGE_SIZE),
        metadata={"description": "Number of results to return", "example": 50},
    )
    offset = fields.Int(
        required=False,
        validate=Range(min=0),
        metadata={"description": "Offset for pagination", "example": 0},
    )
    extra_query = fields.Str(
        required=False,
        validate=INDY_EXTRA_WQL_VALIDATE,
        metadata={
            "description": "(JSON) object mapping referents to extra WQL queries",
            "example": INDY_EXTRA_WQL_EXAMPLE,
        },
    )


class V20PresProblemReportRequestSchema(OpenAPISchema):
    """Request schema for sending problem report."""

    description = fields.Str(required=True)


class V20PresExIdMatchInfoSchema(OpenAPISchema):
    """Path parameters for request taking presentation exchange id."""

    pres_ex_id = fields.Str(
        required=True,
        validate=UUID4_VALIDATE,
        metadata={
            "description": "Presentation exchange identifier",
            "example": UUID4_EXAMPLE,
        },
    )


async def _add_nonce(indy_proof_request: Mapping) -> Mapping:
    """Add nonce to indy proof request if need be."""

    if not indy_proof_request.get("nonce"):
        indy_proof_request["nonce"] = await generate_pr_nonce()
    return indy_proof_request


def _formats_attach(by_format: Mapping, msg_type: str, spec: str) -> Mapping:
    """Break out formats and proposals/requests/presentations for v2.0 messages."""
    attach = []
    for fmt_api, item_by_fmt in by_format.items():
        if (
            fmt_api == V20PresFormat.Format.ANONCREDS.api
            or fmt_api == V20PresFormat.Format.INDY.api
        ):
            attach.append(AttachDecorator.data_base64(mapping=item_by_fmt, ident=fmt_api))
        elif fmt_api == V20PresFormat.Format.DIF.api:
            attach.append(AttachDecorator.data_json(mapping=item_by_fmt, ident=fmt_api))
    return {
        "formats": [
            V20PresFormat(
                attach_id=fmt_api,
                format_=ATTACHMENT_FORMAT[msg_type][fmt_api],
            )
            for fmt_api in by_format
        ],
        f"{spec}_attach": attach,
    }


@docs(tags=["present-proof v2.0"], summary="Fetch all present-proof exchange records")
@querystring_schema(V20PresExRecordListQueryStringSchema)
@response_schema(V20PresExRecordListSchema(), 200, description="")
@tenant_authentication
async def present_proof_list(request: web.BaseRequest):
    """Request handler for searching presentation exchange records.

    Args:
        request: aiohttp request object

    Returns:
        The presentation exchange list response

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    tag_filter = {}
    if "thread_id" in request.query and request.query["thread_id"] != "":
        tag_filter["thread_id"] = request.query["thread_id"]
    post_filter = {
        k: request.query[k]
        for k in ("connection_id", "role", "state")
        if request.query.get(k, "") != ""
    }

    limit, offset, order_by, descending = get_paginated_query_params(request)

    try:
        async with profile.session() as session:
            records = await V20PresExRecord.query(
                session=session,
                tag_filter=tag_filter,
                limit=limit,
                offset=offset,
                order_by=order_by,
                descending=descending,
                post_filter_positive=post_filter,
            )
        results = [record.serialize() for record in records]
    except (StorageError, BaseModelError) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response({"results": results})


@docs(
    tags=["present-proof v2.0"],
    summary="Fetch a single presentation exchange record",
)
@match_info_schema(V20PresExIdMatchInfoSchema())
@response_schema(V20PresExRecordSchema(), 200, description="")
@tenant_authentication
async def present_proof_retrieve(request: web.BaseRequest):
    """Request handler for fetching a single presentation exchange record.

    Args:
        request: aiohttp request object

    Returns:
        The presentation exchange record response

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile
    outbound_handler = request["outbound_message_router"]

    pres_ex_id = request.match_info["pres_ex_id"]
    pres_ex_record = None
    try:
        async with profile.session() as session:
            pres_ex_record = await V20PresExRecord.retrieve_by_id(session, pres_ex_id)
        result = pres_ex_record.serialize()
    except StorageNotFoundError as err:
        # no such pres ex record: not protocol error, user fat-fingered id
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except (BaseModelError, StorageError) as err:
        # present but broken or hopeless: protocol error
        if pres_ex_record:
            async with profile.session() as session:
                await pres_ex_record.save_error_state(session, reason=err.roll_up)
        await report_problem(
            err,
            ProblemReportReason.ABANDONED.value,
            web.HTTPBadRequest,
            pres_ex_record,
            outbound_handler,
        )

    return web.json_response(result)


@docs(
    tags=["present-proof v2.0"],
    summary="Fetch credentials from wallet for presentation request",
)
@match_info_schema(V20PresExIdMatchInfoSchema())
@querystring_schema(V20CredentialsFetchQueryStringSchema())
@response_schema(IndyCredPrecisSchema(many=True), 200, description="")
@tenant_authentication
async def present_proof_credentials_list(request: web.BaseRequest):
    """Request handler for searching applicable credential records.

    Args:
        request: aiohttp request object

    Returns:
        The credential list response

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile
    outbound_handler = request["outbound_message_router"]

    pres_ex_id = request.match_info["pres_ex_id"]
    referents = request.query.get("referent")
    pres_referents = (r.strip() for r in referents.split(",")) if referents else ()

    try:
        async with profile.session() as session:
            pres_ex_record = await V20PresExRecord.retrieve_by_id(session, pres_ex_id)
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err

    # Handle both old style start/count and new limit/offset
    # TODO: Remove start/count and swap to PaginatedQuerySchema and get_limit_offset
    if "limit" in request.query or "offset" in request.query:
        # New style - use limit/offset
        limit = int(request.query.get("limit", DEFAULT_PAGE_SIZE))
        offset = int(request.query.get("offset", 0))
    else:
        # Old style - use start/count
        limit = int(request.query.get("count", "10"))
        offset = int(request.query.get("start", "0"))

    # url encoded json extra_query
    encoded_extra_query = request.query.get("extra_query") or "{}"
    extra_query = json.loads(encoded_extra_query)

    wallet_type = profile.settings.get_value("wallet.type")
    if wallet_type == "askar-anoncreds":
        holder = AnonCredsHolder(profile)
    else:
        holder = profile.inject(IndyHolder)
    credentials = []
    # ANONCREDS or INDY
    try:
        # try anoncreds and fallback to indy
        pres_request = pres_ex_record.by_format["pres_request"].get(
            V20PresFormat.Format.ANONCREDS.api
        )
        if not pres_request:
            pres_request = pres_ex_record.by_format["pres_request"].get(
                V20PresFormat.Format.INDY.api
            )
        if pres_request:
            credentials = (
                await holder.get_credentials_for_presentation_request_by_referent(
                    pres_request,
                    pres_referents,
                    offset=offset,
                    limit=limit,
                    extra_query=extra_query,
                )
            )

    except (IndyHolderError, AnonCredsHolderError) as err:
        if pres_ex_record:
            async with profile.session() as session:
                await pres_ex_record.save_error_state(session, reason=err.roll_up)
        await report_problem(
            err,
            ProblemReportReason.ABANDONED.value,
            web.HTTPBadRequest,
            pres_ex_record,
            outbound_handler,
        )

    dif_holder = profile.inject(VCHolder)
    dif_credentials = []
    dif_cred_value_list = []
    # DIF
    try:
        dif_pres_request = pres_ex_record.by_format["pres_request"].get(
            V20PresFormat.Format.DIF.api
        )
        if dif_pres_request:
            input_descriptors_list = dif_pres_request.get(
                "presentation_definition", {}
            ).get("input_descriptors")
            claim_fmt = dif_pres_request.get("presentation_definition", {}).get("format")
            if claim_fmt and len(claim_fmt.keys()) > 0:
                claim_fmt = ClaimFormat.deserialize(claim_fmt)
            input_descriptors = []
            for input_desc_dict in input_descriptors_list:
                input_descriptors.append(InputDescriptors.deserialize(input_desc_dict))
            record_ids = set()
            for input_descriptor in input_descriptors:
                proof_type = None
                limit_disclosure = input_descriptor.constraint.limit_disclosure and (
                    input_descriptor.constraint.limit_disclosure == "required"
                )
                uri_list = []
                one_of_uri_groups = []
                if input_descriptor.schemas:
                    if input_descriptor.schemas.oneof_filter:
                        one_of_uri_groups = await retrieve_uri_list_from_schema_filter(
                            input_descriptor.schemas.uri_groups
                        )
                    else:
                        schema_uris = input_descriptor.schemas.uri_groups[0]
                        for schema_uri in schema_uris:
                            if schema_uri.required is None:
                                required = True
                            else:
                                required = schema_uri.required
                            if required:
                                uri_list.append(schema_uri.uri)
                if len(uri_list) == 0:
                    uri_list = None
                if len(one_of_uri_groups) == 0:
                    one_of_uri_groups = None
                if limit_disclosure:
                    proof_type = [BbsBlsSignature2020.signature_type]
                if claim_fmt:
                    if claim_fmt.ldp_vp:
                        if "proof_type" in claim_fmt.ldp_vp:
                            proof_types = claim_fmt.ldp_vp.get("proof_type")
                            if limit_disclosure and (
                                BbsBlsSignature2020.signature_type not in proof_types
                            ):
                                raise web.HTTPBadRequest(
                                    reason=(
                                        "Verifier submitted presentation request with "
                                        "limit_disclosure [selective disclosure] "
                                        "option but verifier does not support "
                                        "BbsBlsSignature2020 format"
                                    )
                                )
                            elif (
                                len(proof_types) == 1
                                and (
                                    BbsBlsSignature2020.signature_type not in proof_types
                                )
                                and (
                                    Ed25519Signature2018.signature_type not in proof_types
                                )
                                and (
                                    Ed25519Signature2020.signature_type not in proof_types
                                )
                                and (
                                    EcdsaSecp256r1Signature2019.signature_type
                                    not in proof_types
                                )
                            ):
                                raise web.HTTPBadRequest(
                                    reason=(
                                        "Only BbsBlsSignature2020 and/or "
                                        "Ed25519Signature2018 and/or "
                                        "Ed25519Signature2020 and/or "
                                        "EcdsaSecp256r1Signature2019 signature types "
                                        "are supported"
                                    )
                                )
                            elif (
                                len(proof_types) >= 2
                                and (
                                    BbsBlsSignature2020.signature_type not in proof_types
                                )
                                and (
                                    Ed25519Signature2018.signature_type not in proof_types
                                )
                                and (
                                    Ed25519Signature2020.signature_type not in proof_types
                                )
                                and (
                                    EcdsaSecp256r1Signature2019.signature_type
                                    not in proof_types
                                )
                            ):
                                raise web.HTTPBadRequest(
                                    reason=(
                                        "Only BbsBlsSignature2020, Ed25519Signature2018"
                                        " and Ed25519Signature2020,"
                                        " EcdsaSecp256r1Signature2019"
                                        " signature types are supported"
                                    )
                                )
                            else:
                                for proof_format in proof_types:
                                    if (
                                        proof_format
                                        == Ed25519Signature2018.signature_type
                                    ):
                                        proof_type = [Ed25519Signature2018.signature_type]
                                        break
                                    elif (
                                        proof_format
                                        == Ed25519Signature2020.signature_type
                                    ):
                                        proof_type = [Ed25519Signature2020.signature_type]
                                        break
                                    elif (
                                        proof_format
                                        == EcdsaSecp256r1Signature2019.signature_type
                                    ):
                                        proof_type = [
                                            EcdsaSecp256r1Signature2019.signature_type
                                        ]
                                        break
                                    elif (
                                        proof_format == BbsBlsSignature2020.signature_type
                                    ):
                                        proof_type = [BbsBlsSignature2020.signature_type]
                                        break
                    elif claim_fmt.di_vc:
                        if "proof_type" in claim_fmt.di_vc:
                            proof_types = claim_fmt.di_vc.get("proof_type")

                            proof_type = [
                                "DataIntegrityProof"
                            ]  # [LinkedDataProof.signature_type]

                        # TODO check acceptable proof type(s) ("anoncreds-2023")
                    else:
                        raise web.HTTPBadRequest(
                            reason=(
                                "Currently, only ldp_vp with "
                                "BbsBlsSignature2020, Ed25519Signature2018 and "
                                "Ed25519Signature2020, EcdsaSecp256r1Signature2019"
                                " signature types are supported"
                            )
                        )
                if one_of_uri_groups:
                    records = []
                    cred_group_record_ids = set()
                    for uri_group in one_of_uri_groups:
                        search = dif_holder.search_credentials(
                            proof_types=proof_type, pd_uri_list=uri_group
                        )
                        cred_group = await search.fetch(limit)
                        (
                            cred_group_vcrecord_list,
                            cred_group_vcrecord_ids_set,
                        ) = await process_vcrecords_return_list(
                            cred_group, cred_group_record_ids
                        )
                        cred_group_record_ids = cred_group_vcrecord_ids_set
                        records = records + cred_group_vcrecord_list
                else:
                    search = dif_holder.search_credentials(
                        proof_types=proof_type,
                        pd_uri_list=uri_list,
                    )
                    records = await search.fetch(limit)
                # Avoiding addition of duplicate records
                vcrecord_list, vcrecord_ids_set = await process_vcrecords_return_list(
                    records, record_ids
                )
                record_ids = vcrecord_ids_set
                dif_credentials = dif_credentials + vcrecord_list
            for dif_credential in dif_credentials:
                cred_value = dif_credential.cred_value
                cred_value["record_id"] = dif_credential.record_id
                dif_cred_value_list.append(cred_value)
    except (
        StorageNotFoundError,
        V20PresFormatHandlerError,
    ) as err:
        if pres_ex_record:
            async with profile.session() as session:
                await pres_ex_record.save_error_state(session, reason=err.roll_up)
        await report_problem(
            err,
            ProblemReportReason.ABANDONED.value,
            web.HTTPBadRequest,
            pres_ex_record,
            outbound_handler,
        )
    credentials = list(credentials) + dif_cred_value_list
    return web.json_response(credentials)


async def process_vcrecords_return_list(
    vc_records: Sequence[VCRecord], record_ids: set
) -> Tuple[Sequence[VCRecord], set]:
    """Return list of non-duplicate VCRecords."""
    to_add = []
    for vc_record in vc_records:
        if vc_record.record_id not in record_ids:
            to_add.append(vc_record)
            record_ids.add(vc_record.record_id)
    return (to_add, record_ids)


async def retrieve_uri_list_from_schema_filter(
    schema_uri_groups: Sequence[Sequence[SchemaInputDescriptor]],
) -> Sequence[str]:
    """Retrieve list of schema uri from uri_group."""
    group_schema_uri_list = []
    for schema_group in schema_uri_groups:
        uri_list = []
        for schema in schema_group:
            uri_list.append(schema.uri)
        if len(uri_list) > 0:
            group_schema_uri_list.append(uri_list)
    return group_schema_uri_list


@docs(tags=["present-proof v2.0"], summary="Sends a presentation proposal")
@request_schema(V20PresProposalRequestSchema())
@response_schema(V20PresExRecordSchema(), 200, description="")
@tenant_authentication
async def present_proof_send_proposal(request: web.BaseRequest):
    """Request handler for sending a presentation proposal.

    Args:
        request: aiohttp request object

    Returns:
        The presentation exchange details

    """
    r_time = get_timer()

    context: AdminRequestContext = request["context"]
    profile = context.profile
    outbound_handler = request["outbound_message_router"]

    body = await request.json()

    comment = body.get("comment")
    connection_id = body.get("connection_id")

    pres_proposal = body.get("presentation_proposal")
    conn_record = None
    try:
        async with profile.session() as session:
            conn_record = await ConnRecord.retrieve_by_id(session, connection_id)
        pres_proposal_message = V20PresProposal(
            comment=comment,
            **_formats_attach(pres_proposal, PRES_20_PROPOSAL, "proposals"),
        )
    except (BaseModelError, StorageError) as err:
        # other party does not care about our false protocol start
        raise web.HTTPBadRequest(reason=err.roll_up)

    if not conn_record.is_ready:
        raise web.HTTPForbidden(reason=f"Connection {connection_id} not ready")

    trace_msg = body.get("trace")
    pres_proposal_message.assign_trace_decorator(
        context.settings,
        trace_msg,
    )
    auto_present = body.get(
        "auto_present", context.settings.get("debug.auto_respond_presentation_request")
    )
    auto_remove = body.get("auto_remove")

    pres_manager = V20PresManager(profile)
    pres_ex_record = None
    try:
        pres_ex_record = await pres_manager.create_exchange_for_proposal(
            connection_id=connection_id,
            pres_proposal_message=pres_proposal_message,
            auto_present=auto_present,
            auto_remove=auto_remove,
        )
        result = pres_ex_record.serialize()
    except (BaseModelError, StorageError) as err:
        if pres_ex_record:
            async with profile.session() as session:
                await pres_ex_record.save_error_state(session, reason=err.roll_up)
        # other party does not care about our false protocol start
        raise web.HTTPBadRequest(reason=err.roll_up)

    await outbound_handler(pres_proposal_message, connection_id=connection_id)

    trace_event(
        context.settings,
        pres_proposal_message,
        outcome="presentation_exchange_propose.END",
        perf_counter=r_time,
    )

    return web.json_response(result)


@docs(
    tags=["present-proof v2.0"],
    summary="Creates a presentation request not bound to any proposal or connection",
)
@request_schema(V20PresCreateRequestRequestSchema())
@response_schema(V20PresExRecordSchema(), 200, description="")
@tenant_authentication
async def present_proof_create_request(request: web.BaseRequest):
    """Request handler for creating a free presentation request.

    The presentation request will not be bound to any proposal
    or existing connection.

    Args:
        request: aiohttp request object

    Returns:
        The presentation exchange details

    """
    r_time = get_timer()

    context: AdminRequestContext = request["context"]
    profile = context.profile
    outbound_handler = request["outbound_message_router"]

    body = await request.json()

    comment = body.get("comment")
    pres_request_spec = body.get("presentation_request")
    if pres_request_spec:
        if V20PresFormat.Format.INDY.api in pres_request_spec:
            await _add_nonce(pres_request_spec[V20PresFormat.Format.INDY.api])
        if V20PresFormat.Format.ANONCREDS.api in pres_request_spec:
            await _add_nonce(pres_request_spec[V20PresFormat.Format.ANONCREDS.api])

    pres_request_message = V20PresRequest(
        comment=comment,
        will_confirm=True,
        **_formats_attach(pres_request_spec, PRES_20_REQUEST, "request_presentations"),
    )
    auto_verify = body.get(
        "auto_verify", context.settings.get("debug.auto_verify_presentation")
    )
    auto_remove = body.get("auto_remove")
    trace_msg = body.get("trace")
    pres_request_message.assign_trace_decorator(
        context.settings,
        trace_msg,
    )

    pres_manager = V20PresManager(profile)
    pres_ex_record = None
    try:
        pres_ex_record = await pres_manager.create_exchange_for_request(
            connection_id=None,
            pres_request_message=pres_request_message,
            auto_verify=auto_verify,
            auto_remove=auto_remove,
        )
        result = pres_ex_record.serialize()
    except (BaseModelError, StorageError) as err:
        if pres_ex_record:
            async with profile.session() as session:
                await pres_ex_record.save_error_state(session, reason=err.roll_up)
        # other party does not care about our false protocol start
        raise web.HTTPBadRequest(reason=err.roll_up)

    await outbound_handler(pres_request_message, connection_id=None)

    trace_event(
        context.settings,
        pres_request_message,
        outcome="presentation_exchange_create_request.END",
        perf_counter=r_time,
    )

    return web.json_response(result)


@docs(
    tags=["present-proof v2.0"],
    summary="Sends a free presentation request not bound to any proposal",
)
@request_schema(V20PresSendRequestRequestSchema())
@response_schema(V20PresExRecordSchema(), 200, description="")
@tenant_authentication
async def present_proof_send_free_request(request: web.BaseRequest):
    """Request handler for sending a presentation request free from any proposal.

    Args:
        request: aiohttp request object

    Returns:
        The presentation exchange details

    """
    r_time = get_timer()

    context: AdminRequestContext = request["context"]
    profile = context.profile
    outbound_handler = request["outbound_message_router"]

    body = await request.json()

    connection_id = body.get("connection_id")
    try:
        async with profile.session() as session:
            conn_record = await ConnRecord.retrieve_by_id(session, connection_id)
    except StorageNotFoundError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    if not conn_record.is_ready:
        raise web.HTTPForbidden(reason=f"Connection {connection_id} not ready")

    comment = body.get("comment")
    pres_request_spec = body.get("presentation_request")
    if pres_request_spec:
        if V20PresFormat.Format.INDY.api in pres_request_spec:
            await _add_nonce(pres_request_spec[V20PresFormat.Format.INDY.api])
        if V20PresFormat.Format.ANONCREDS.api in pres_request_spec:
            await _add_nonce(pres_request_spec[V20PresFormat.Format.ANONCREDS.api])
    pres_request_message = V20PresRequest(
        comment=comment,
        will_confirm=True,
        **_formats_attach(pres_request_spec, PRES_20_REQUEST, "request_presentations"),
    )
    auto_verify = body.get(
        "auto_verify", context.settings.get("debug.auto_verify_presentation")
    )
    auto_remove = body.get("auto_remove")
    trace_msg = body.get("trace")
    pres_request_message.assign_trace_decorator(
        context.settings,
        trace_msg,
    )

    pres_manager = V20PresManager(profile)
    pres_ex_record = None
    try:
        pres_ex_record = await pres_manager.create_exchange_for_request(
            connection_id=connection_id,
            pres_request_message=pres_request_message,
            auto_verify=auto_verify,
            auto_remove=auto_remove,
        )
        result = pres_ex_record.serialize()
    except (BaseModelError, StorageError) as err:
        if pres_ex_record:
            async with profile.session() as session:
                await pres_ex_record.save_error_state(session, reason=err.roll_up)
        # other party does not care about our false protocol start
        raise web.HTTPBadRequest(reason=err.roll_up)

    await outbound_handler(pres_request_message, connection_id=connection_id)

    trace_event(
        context.settings,
        pres_request_message,
        outcome="presentation_exchange_send_request.END",
        perf_counter=r_time,
    )

    return web.json_response(result)


@docs(
    tags=["present-proof v2.0"],
    summary="Sends a presentation request in reference to a proposal",
)
@match_info_schema(V20PresExIdMatchInfoSchema())
@request_schema(V20PresentationSendRequestToProposalSchema())
@response_schema(V20PresExRecordSchema(), 200, description="")
@tenant_authentication
async def present_proof_send_bound_request(request: web.BaseRequest):
    """Request handler for sending a presentation request bound to a proposal.

    Args:
        request: aiohttp request object

    Returns:
        The presentation exchange details

    """
    r_time = get_timer()

    context: AdminRequestContext = request["context"]
    profile = context.profile
    outbound_handler = request["outbound_message_router"]

    body = await request.json()

    pres_ex_id = request.match_info["pres_ex_id"]
    pres_ex_record = None
    try:
        async with profile.session() as session:
            pres_ex_record = await V20PresExRecord.retrieve_by_id(session, pres_ex_id)
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err

    if pres_ex_record.state != (V20PresExRecord.STATE_PROPOSAL_RECEIVED):
        raise web.HTTPBadRequest(
            reason=(
                f"Presentation exchange {pres_ex_id} "
                f"in {pres_ex_record.state} state "
                f"(must be {V20PresExRecord.STATE_PROPOSAL_RECEIVED})"
            )
        )
    connection_id = pres_ex_record.connection_id

    try:
        async with profile.session() as session:
            conn_record = await ConnRecord.retrieve_by_id(session, connection_id)
    except StorageError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    if not conn_record.is_ready:
        raise web.HTTPForbidden(reason=f"Connection {connection_id} not ready")

    pres_ex_record.auto_verify = body.get(
        "auto_verify", context.settings.get("debug.auto_verify_presentation")
    )
    pres_ex_record.auto_remove = body.get("auto_remove")
    pres_manager = V20PresManager(profile)
    try:
        (
            pres_ex_record,
            pres_request_message,
        ) = await pres_manager.create_bound_request(pres_ex_record)
        result = pres_ex_record.serialize()
    except (BaseModelError, LedgerError, StorageError) as err:
        if pres_ex_record:
            async with profile.session() as session:
                await pres_ex_record.save_error_state(session, reason=err.roll_up)
        # other party cares that we cannot continue protocol
        await report_problem(
            err,
            ProblemReportReason.ABANDONED.value,
            web.HTTPBadRequest,
            pres_ex_record,
            outbound_handler,
        )

    trace_msg = body.get("trace")
    pres_request_message.assign_trace_decorator(
        context.settings,
        trace_msg,
    )
    await outbound_handler(pres_request_message, connection_id=connection_id)

    trace_event(
        context.settings,
        pres_request_message,
        outcome="presentation_exchange_send_request.END",
        perf_counter=r_time,
    )

    return web.json_response(result)


@docs(tags=["present-proof v2.0"], summary="Sends a proof presentation")
@match_info_schema(V20PresExIdMatchInfoSchema())
@request_schema(V20PresSpecByFormatRequestSchema())
@response_schema(V20PresExRecordSchema(), description="")
@tenant_authentication
async def present_proof_send_presentation(request: web.BaseRequest):
    """Request handler for sending a presentation.

    Args:
        request: aiohttp request object

    Returns:
        The presentation exchange details

    """
    r_time = get_timer()

    context: AdminRequestContext = request["context"]
    profile = context.profile
    outbound_handler = request["outbound_message_router"]
    pres_ex_id = request.match_info["pres_ex_id"]
    body = await request.json()
    supported_formats = ["anoncreds", "dif", "indy"]
    if not any(x in body for x in supported_formats):
        raise web.HTTPBadRequest(
            reason=(
                "No presentation format specification provided, "
                "either dif, anoncreds or indy must be included. "
                "In case of DIF, if no additional specification "
                'needs to be provided then include "dif": {}'
            )
        )
    comment = body.get("comment")
    pres_ex_record = None
    try:
        async with profile.session() as session:
            pres_ex_record = await V20PresExRecord.retrieve_by_id(session, pres_ex_id)
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err

    if pres_ex_record.state != (V20PresExRecord.STATE_REQUEST_RECEIVED):
        raise web.HTTPBadRequest(
            reason=(
                f"Presentation exchange {pres_ex_id} "
                f"in {pres_ex_record.state} state "
                f"(must be {V20PresExRecord.STATE_REQUEST_RECEIVED})"
            )
        )

    auto_remove = body.get("auto_remove")
    if auto_remove is None:
        auto_remove = not profile.settings.get("preserve_exchange_records")

    pres_ex_record.auto_remove = auto_remove

    # Fetch connection if exchange has record
    conn_record = None
    if pres_ex_record.connection_id:
        try:
            async with profile.session() as session:
                conn_record = await ConnRecord.retrieve_by_id(
                    session, pres_ex_record.connection_id
                )
        except StorageNotFoundError as err:
            raise web.HTTPBadRequest(reason=err.roll_up) from err

    if conn_record and not conn_record.is_ready:
        raise web.HTTPForbidden(
            reason=f"Connection {pres_ex_record.connection_id} not ready"
        )

    pres_manager = V20PresManager(profile)
    try:
        pres_ex_record, pres_message = await pres_manager.create_pres(
            pres_ex_record,
            request_data=body,
            comment=comment,
        )
        result = pres_ex_record.serialize()
    except (
        BaseModelError,
        IndyHolderError,
        AnonCredsHolderError,
        LedgerError,
        V20PresFormatHandlerError,
        StorageError,
        WalletNotFoundError,
    ) as err:
        async with profile.session() as session:
            await pres_ex_record.save_error_state(session, reason=err.roll_up)
        # other party cares that we cannot continue protocol
        await report_problem(
            err,
            ProblemReportReason.ABANDONED.value,
            web.HTTPBadRequest,
            pres_ex_record,
            outbound_handler,
        )
    trace_msg = body.get("trace")
    pres_message.assign_trace_decorator(
        context.settings,
        trace_msg,
    )
    await outbound_handler(pres_message, connection_id=pres_ex_record.connection_id)

    trace_event(
        context.settings,
        pres_message,
        outcome="presentation_exchange_send_request.END",
        perf_counter=r_time,
    )

    return web.json_response(result)


@docs(tags=["present-proof v2.0"], summary="Verify a received presentation")
@match_info_schema(V20PresExIdMatchInfoSchema())
@response_schema(V20PresExRecordSchema(), description="")
@tenant_authentication
async def present_proof_verify_presentation(request: web.BaseRequest):
    """Request handler for verifying a presentation request.

    Args:
        request: aiohttp request object

    Returns:
        The presentation exchange details

    """
    r_time = get_timer()

    context: AdminRequestContext = request["context"]
    profile = context.profile
    outbound_handler = request["outbound_message_router"]

    pres_ex_id = request.match_info["pres_ex_id"]

    pres_ex_record = None
    try:
        async with profile.session() as session:
            pres_ex_record = await V20PresExRecord.retrieve_by_id(session, pres_ex_id)
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err

    if pres_ex_record.state != (V20PresExRecord.STATE_PRESENTATION_RECEIVED):
        raise web.HTTPBadRequest(
            reason=(
                f"Presentation exchange {pres_ex_id} "
                f"in {pres_ex_record.state} state "
                f"(must be {V20PresExRecord.STATE_PRESENTATION_RECEIVED})"
            )
        )

    pres_manager = V20PresManager(profile)
    try:
        pres_ex_record = await pres_manager.verify_pres(pres_ex_record)
        result = pres_ex_record.serialize()
    except (BaseModelError, LedgerError, StorageError) as err:
        if pres_ex_record:
            async with profile.session() as session:
                await pres_ex_record.save_error_state(session, reason=err.roll_up)
        # other party cares that we cannot continue protocol
        await report_problem(
            err,
            ProblemReportReason.ABANDONED.value,
            web.HTTPBadRequest,
            pres_ex_record,
            outbound_handler,
        )

    trace_event(
        context.settings,
        pres_ex_record,
        outcome="presentation_exchange_verify.END",
        perf_counter=r_time,
    )

    return web.json_response(result)


@docs(
    tags=["present-proof v2.0"],
    summary="Send a problem report for presentation exchange",
)
@match_info_schema(V20PresExIdMatchInfoSchema())
@request_schema(V20PresProblemReportRequestSchema())
@response_schema(V20PresentProofModuleResponseSchema(), 200, description="")
@tenant_authentication
async def present_proof_problem_report(request: web.BaseRequest):
    """Request handler for sending problem report.

    Args:
        request: aiohttp request object

    """
    context: AdminRequestContext = request["context"]
    outbound_handler = request["outbound_message_router"]

    pres_ex_id = request.match_info["pres_ex_id"]
    body = await request.json()
    description = body["description"]

    try:
        async with context.profile.session() as session:
            pres_ex_record = await V20PresExRecord.retrieve_by_id(session, pres_ex_id)
            await pres_ex_record.save_error_state(
                session,
                reason=f"created problem report: {description}",
            )
        report = problem_report_for_record(pres_ex_record, description)
    except StorageNotFoundError as err:  # other party does not care about meta-problems
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except StorageError as err:
        raise web.HTTPBadRequest(reason=err.roll_up)

    await outbound_handler(report, connection_id=pres_ex_record.connection_id)

    return web.json_response({})


@docs(
    tags=["present-proof v2.0"],
    summary="Remove an existing presentation exchange record",
)
@match_info_schema(V20PresExIdMatchInfoSchema())
@response_schema(V20PresentProofModuleResponseSchema(), description="")
@tenant_authentication
async def present_proof_remove(request: web.BaseRequest):
    """Request handler for removing a presentation exchange record.

    Args:
        request: aiohttp request object

    """
    context: AdminRequestContext = request["context"]

    pres_ex_id = request.match_info["pres_ex_id"]
    pres_ex_record = None
    try:
        async with context.profile.session() as session:
            try:
                pres_ex_record = await V20PresExRecord.retrieve_by_id(session, pres_ex_id)
                await pres_ex_record.delete_record(session)
            except (BaseModelError, ValidationError):
                storage = session.inject(BaseStorage)
                storage_record = await storage.get_record(
                    record_type=V20PresExRecord.RECORD_TYPE, record_id=pres_ex_id
                )
                await storage.delete_record(storage_record)
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except StorageError as err:
        raise web.HTTPBadRequest(reason=err.roll_up)

    return web.json_response({})


async def register(app: web.Application):
    """Register routes."""

    app.add_routes(
        [
            web.get(
                "/present-proof-2.0/records",
                present_proof_list,
                allow_head=False,
            ),
            web.get(
                "/present-proof-2.0/records/{pres_ex_id}",
                present_proof_retrieve,
                allow_head=False,
            ),
            web.get(
                "/present-proof-2.0/records/{pres_ex_id}/credentials",
                present_proof_credentials_list,
                allow_head=False,
            ),
            web.post(
                "/present-proof-2.0/send-proposal",
                present_proof_send_proposal,
            ),
            web.post(
                "/present-proof-2.0/create-request",
                present_proof_create_request,
            ),
            web.post(
                "/present-proof-2.0/send-request",
                present_proof_send_free_request,
            ),
            web.post(
                "/present-proof-2.0/records/{pres_ex_id}/send-request",
                present_proof_send_bound_request,
            ),
            web.post(
                "/present-proof-2.0/records/{pres_ex_id}/send-presentation",
                present_proof_send_presentation,
            ),
            web.post(
                "/present-proof-2.0/records/{pres_ex_id}/verify-presentation",
                present_proof_verify_presentation,
            ),
            web.post(
                "/present-proof-2.0/records/{pres_ex_id}/problem-report",
                present_proof_problem_report,
            ),
            web.delete(
                "/present-proof-2.0/records/{pres_ex_id}",
                present_proof_remove,
            ),
        ]
    )


def post_process_routes(app: web.Application):
    """Amend swagger API."""

    # Add top-level tags description
    if "tags" not in app._state["swagger_dict"]:
        app._state["swagger_dict"]["tags"] = []
    app._state["swagger_dict"]["tags"].append(
        {
            "name": "present-proof v2.0",
            "description": "Proof presentation v2.0",
            "externalDocs": {"description": "Specification", "url": SPEC_URI},
        }
    )



TEST_SEED = "testseed000000000000000000000001"
TEST_DID = "55GkHamhTU1ZbTbV2ab9DE"
TEST_VERKEY = "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx"

TEST_SIGN_OBJ0 = {
    "doc": {
        "@context": [
            "https://www.w3.org/2018/credentials/v1",
            "https://www.w3.org/2018/credentials/examples/v1",
        ],
        "id": "http://example.gov/credentials/3732",
        "type": ["VerifiableCredential", "UniversityDegreeCredential"],
        "issuer": ("did:key:z6MkjRagNiMu91DduvCvgEsqLZDVzrJzFrwahc4tXLt9DoHd"),
        "issuanceDate": "2020-03-10T04:24:12.164Z",
        "credentialSubject": {
            "id": ("did:key:z6MkjRagNiMu91DduvCvgEsqLZDVzrJzFrwahc4tXLt9DoHd"),
            "degree": {
                "type": "BachelorDegree",
                "name": "Bachelor of Science and Arts",
            },
        },
    },
    "options": {
        "verificationMethod": "did:example:123#z6MksHh7qHWvybLg5QTPPdG2DgEjjduBDArV9EF9mRiRzMBN",
        "proofPurpose": "assertionMethod",
        "created": "2020-04-02T18:48:36Z",
        "domain": "example.com",
        "challenge": "d436f0c8-fbd9-4e48-bbb2-55fc5d0920a8",
    },
}
TEST_SIGN_OBJ1 = {
    "doc": {
        "@context": [
            "https://www.w3.org/2018/credentials/v1",
            "https://www.w3.org/2018/credentials/examples/v1",
        ],
        "id": "http://example.gov/credentials/3735",
        "type": ["VerifiableCredential", "UniversityDegreeCredential"],
        "issuer": "did:example:123",
        "issuanceDate": "2020-03-16T22:37:26.544Z",
        "credentialSubject": {
            "id": "did:example:123",
            "degree": {
                "type": "BachelorDegree",
                "name": "Bachelor of Science and Arts",
            },
        },
    },
    "options": {
        "verificationMethod": "did:example:123#z6MksHh7qHWvybLg5QTPPdG2DgEjjduBDArV9EF9mRiRzMBN",
        "proofPurpose": "assertionMethod",
        "created": "2020-04-02T18:48:36Z",
        "domain": "example.com",
        "challenge": "d436f0c8-fbd9-4e48-bbb2-55fc5d0920a8",
    },
}
TEST_SIGN_OBJ2 = {
    "doc": {
        "@context": [
            "https://www.w3.org/2018/credentials/v1",
            "https://www.w3.org/2018/credentials/examples/v1",
        ],
        "holder": "did:example:123",
        "type": "VerifiablePresentation",
        "verifiableCredential": [
            {
                "@context": [
                    "https://www.w3.org/2018/credentials/v1",
                    "https://www.w3.org/2018/credentials/examples/v1",
                ]
            },
            {"id": "http://example.gov/credentials/3732"},
            {"type": ["VerifiableCredential", "UniversityDegreeCredential"]},
            {"issuer": "did:example:123"},
            {"issuanceDate": "2020-03-16T22:37:26.544Z"},
            {
                "credentialSubject": {
                    "id": "did:example:123",
                    "degree": {
                        "type": "BachelorDegree",
                        "name": "Bachelor of Science and Arts",
                    },
                }
            },
            {
                "proof": {
                    "type": "Ed25519Signature2018",
                    "created": "2020-04-02T18:28:08Z",
                    "verificationMethod": "did:example:123#z6MksHh7qHWvybLg5QTPPdG2DgEjjduBDArV9EF9mRiRzMBN",
                    "proofPurpose": "assertionMethod",
                    "jws": "eyJhbGciOiJFZERTQSIsImI2NCI6ZmFsc2UsImNyaXQiOlsiYjY0Il19..YtqjEYnFENT7fNW-COD0HAACxeuQxPKAmp4nIl8jYAu__6IH2FpSxv81w-l5PvE1og50tS9tH8WyXMlXyo45CA",
                }
            },
        ],
        "proof": {
            "verificationMethod": "did:example:123#z6MksHh7qHWvybLg5QTPPdG2DgEjjduBDArV9EF9mRiRzMBN",
            "proofPurpose": "assertionMethod",
            "created": "2020-04-02T18:48:36Z",
            "domain": "example.com",
            "challenge": "d436f0c8-fbd9-4e48-bbb2-55fc5d0920a8",
            "type": "Ed25519Signature2018",
            "jws": "eyJhbGciOiAiRWREU0EiLCAiYjY0IjogZmFsc2UsICJjcml0IjogWyJiNjQiXX0..a6dB9OAI9HWc1lDoWzd1---XF_QdArVMu99N2OKnOFT2Ize8MiuVvbJCIkYHpjn3arPle-o0iMlUx3q08ES_Bg",
        },
    },
    "options": {
        "verificationMethod": "did:example:123#z6MksHh7qHWvybLg5QTPPdG2DgEjjduBDArV9EF9mRiRzMBN",
        "proofPurpose": "assertionMethod",
        "created": "2020-04-02T18:48:36Z",
        "domain": "example.com",
        "challenge": "d436f0c8-fbd9-4e48-bbb2-55fc5d0920a8",
    },
}

TEST_SIGN_OBJS = [TEST_SIGN_OBJ0, TEST_SIGN_OBJ1, TEST_SIGN_OBJ2]  # , TEST_SIGN_OBJ3]

TEST_SIGN_ERROR_OBJ0 = {
    "doc": {
        "@context": [
            "https://www.w3.org/2018/credentials/v1",
            "https://essif-lab.pages.grnet.gr/interoperability/eidas-generic-use-case/contexts/ehic-v1.jsonld",
        ],
        "id": "https://ec.europa.eu/credentials/83627465",
        "type": ["VerifiableCredential", "EuropeanHealthInsuranceCard"],
        "issuer": "did:example:28394728934792387",
        "name": "European Health Insurance Card",
        "description": "Example of a European Health Insurance Card",
        "attribute2drop": "drop it like it's hot",
        "issuanceDate": "2021-01-01T12:19:52Z",
        "expirationDate": "2029-12-03T12:19:52Z",
        "institutionID": "09999 - GE KVG",
        "cardNo": "80756099990000034111",
        "personalID": "09999 111999",
        "credentialSubject": {
            "id": "did:example:b34ca6cd37bbf23",
            "type": ["EuropeanHealthInsuranceHolder", "Person"],
            "familyName": "Muster",
            "giveName": "Maria",
            "birthDate": "1958-07-17",
        },
    },
    "options": {
        "verificationMethod": "did:example:123#z6MksHh7qHWvybLg5QTPPdG2DgEjjduBDArV9EF9mRiRzMBN",
        "proofPurpose": "assertionMethod",
        "created": "2020-04-02T18:48:36Z",
        "domain": "example.com",
        "challenge": "d436f0c8-fbd9-4e48-bbb2-55fc5d0920a8",
    },
}
TEST_SIGN_ERROR_OBJ1 = {
    "doc": {
        "@context": [
            "https://www.w3.org/2018/credentials/v1",
            "https://essif-lab.pages.grnet.gr/interoperability/eidas-generic-use-case/contexts/ehic-v1.jsonld",
        ],
        "id": "https://ec.europa.eu/credentials/83627465",
        "type": ["VerifiableCredential", "EuropeanHealthInsuranceCard"],
        "issuer": "did:example:28394728934792387",
        "name": "European Health Insurance Card",
        "description": "Example of a European Health Insurance Card",
        "issuanceDate": "2021-01-01T12:19:52Z",
        "expirationDate": "2029-12-03T12:19:52Z",
        "institutionID": "09999 - GE KVG",
        "cardNo": "80756099990000034111",
        "personalID": "09999 111999",
        "credentialSubject": {
            "id": "did:example:b34ca6cd37bbf23",
            "type": ["EuropeanHealthInsuranceHolder", "Person"],
            "attribute2drop": "drop it like it's hot",
            "familyName": "Muster",
            "giveName": "Maria",
            "birthDate": "1958-07-17",
        },
    },
    "options": {
        "verificationMethod": "did:example:123#z6MksHh7qHWvybLg5QTPPdG2DgEjjduBDArV9EF9mRiRzMBN",
        "proofPurpose": "assertionMethod",
        "created": "2020-04-02T18:48:36Z",
        "domain": "example.com",
        "challenge": "d436f0c8-fbd9-4e48-bbb2-55fc5d0920a8",
    },
}
TEST_VALIDATE_ERROR_OBJ2 = {
    "doc": {
        "@context": [
            "https://www.w3.org/2018/credentials/v1",
            "https://www.w3.org/2018/credentials/examples/v1",
        ],
        "holder": "did:example:123",
        "type": "VerifiablePresentation",
        "verifiableCredential": [
            {
                "@context": [
                    "https://www.w3.org/2018/credentials/v1",
                    "https://www.w3.org/2018/credentials/examples/v1",
                ]
            },
            {"id": "http://example.gov/credentials/3732"},
            {"type": ["VerifiableCredential", "UniversityDegreeCredential"]},
            {"issuer": "did:example:123"},
            {"issuanceDate": "2020-03-16T22:37:26.544Z"},
            {
                "credentialSubject": {
                    "id": "did:example:123",
                    "degree": {
                        "type": "BachelorDegree",
                        "name": "Bachelor of Science and Arts",
                    },
                }
            },
            {
                "proof": {
                    "type": "Ed25519Signature2018",
                    "created": "2020-04-02T18:28:08Z",
                    "verificationMethod": "did:example:123#z6MksHh7qHWvybLg5QTPPdG2DgEjjduBDArV9EF9mRiRzMBN",
                    "proofPurpose": "assertionMethod",
                    "jws": "eyJhbGciOiJFZERTQSIsImI2NCI6ZmFsc2UsImNyaXQiOlsiYjY0Il19..YtqjEYnFENT7fNW-COD0HAACxeuQxPKAmp4nIl8jYAu__6IH2FpSxv81w-l5PvE1og50tS9tH8WyXMlXyo45CA",
                }
            },
        ],
        "proof": {
            "verificationMethod": "did:example:123#z6MksHh7qHWvybLg5QTPPdG2DgEjjduBDArV9EF9mRiRzMBN",
            "attribute2drop": "drop it like it's hot",
            "proofPurpose": "assertionMethod",
            "created": "2020-04-02T18:48:36Z",
            "domain": "example.com",
            "challenge": "d436f0c8-fbd9-4e48-bbb2-55fc5d0920a8",
            "type": "Ed25519Signature2018",
            "jws": "eyJhbGciOiAiRWREU0EiLCAiYjY0IjogZmFsc2UsICJjcml0IjogWyJiNjQiXX0..a6dB9OAI9HWc1lDoWzd1---XF_QdArVMu99N2OKnOFT2Ize8MiuVvbJCIkYHpjn3arPle-o0iMlUx3q08ES_Bg",
        },
    },
    "verkey": "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx",
}
TEST_SIGN_ERROR_OBJS = [TEST_SIGN_ERROR_OBJ0, TEST_SIGN_ERROR_OBJ1]

TEST_VERIFY_OBJ0 = {
    "verkey": ("5yKdnU7ToTjAoRNDzfuzVTfWBH38qyhE1b9xh4v8JaWF"),
    "doc": {
        "@context": [
            "https://www.w3.org/2018/credentials/v1",
            "https://www.w3.org/2018/credentials/examples/v1",
        ],
        "id": "http://example.gov/credentials/3732",
        "type": ["VerifiableCredential", "UniversityDegreeCredential"],
        "issuer": ("did:key:z6MkjRagNiMu91DduvCvgEsqLZDVzrJzFrwahc4tXLt9DoHd"),
        "issuanceDate": "2020-03-10T04:24:12.164Z",
        "credentialSubject": {
            "id": ("did:key:z6MkjRagNiMu91DduvCvgEsqLZDVzrJzFrwahc4tXLt9DoHd"),
            "degree": {
                "type": "BachelorDegree",
                "name": "Bachelor of Science and Arts",
            },
        },
        "proof": {
            "type": "Ed25519Signature2018",
            "created": "2020-04-10T21:35:35Z",
            "verificationMethod": (
                "did:key:"
                "z6MkjRagNiMu91DduvCvgEsqLZDVzrJzFrwahc"
                "4tXLt9DoHd#z6MkjRagNiMu91DduvCvgEsqLZD"
                "VzrJzFrwahc4tXLt9DoHd"
            ),
            "proofPurpose": "assertionMethod",
            "jws": (
                "eyJhbGciOiJFZERTQSIsImI2NCI6ZmFsc2UsImNyaX"
                "QiOlsiYjY0Il19..l9d0YHjcFAH2H4dB9xlWFZQLUp"
                "ixVCWJk0eOt4CXQe1NXKWZwmhmn9OQp6YxX0a2Lffe"
                "gtYESTCJEoGVXLqWAA"
            ),
        },
    },
}
TEST_VERIFY_OBJ1 = {
    "doc": {
        "@context": [
            "https://www.w3.org/2018/credentials/v1",
            "https://www.w3.org/2018/credentials/examples/v1",
        ],
        "id": "http://example.gov/credentials/3732",
        "type": ["VerifiableCredential", "UniversityDegreeCredential"],
        "issuer": "did:example:123",
        "issuanceDate": "2020-03-16T22:37:26.544Z",
        "credentialSubject": {
            "id": "did:example:123",
            "degree": {
                "type": "BachelorDegree",
                "name": "Bachelor of Science and Arts",
            },
        },
        "proof": {
            "verificationMethod": "did:example:123#z6MksHh7qHWvybLg5QTPPdG2DgEjjduBDArV9EF9mRiRzMBN",
            "proofPurpose": "assertionMethod",
            "created": "2020-04-02T18:48:36Z",
            "domain": "example.com",
            "challenge": "d436f0c8-fbd9-4e48-bbb2-55fc5d0920a8",
            "type": "Ed25519Signature2018",
            "jws": "eyJhbGciOiAiRWREU0EiLCAiYjY0IjogZmFsc2UsICJjcml0IjogWyJiNjQiXX0..MthZGAH62bEu2e4rZSE6b0XvGr_5z6J3FSXuVJnOOxr6sgdJpUenXJ-113MTtjArwC2JXh0zeolhXithxud_Dw",
        },
    },
    "verkey": "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx",
}
TEST_VERIFY_OBJ2 = {
    "doc": {
        "@context": [
            "https://www.w3.org/2018/credentials/v1",
            "https://www.w3.org/2018/credentials/examples/v1",
        ],
        "holder": "did:example:123",
        "type": "VerifiablePresentation",
        "verifiableCredential": [
            {
                "@context": [
                    "https://www.w3.org/2018/credentials/v1",
                    "https://www.w3.org/2018/credentials/examples/v1",
                ]
            },
            {"id": "http://example.gov/credentials/3732"},
            {"type": ["VerifiableCredential", "UniversityDegreeCredential"]},
            {"issuer": "did:example:123"},
            {"issuanceDate": "2020-03-16T22:37:26.544Z"},
            {
                "credentialSubject": {
                    "id": "did:example:123",
                    "degree": {
                        "type": "BachelorDegree",
                        "name": "Bachelor of Science and Arts",
                    },
                }
            },
            {
                "proof": {
                    "type": "Ed25519Signature2018",
                    "created": "2020-04-02T18:28:08Z",
                    "verificationMethod": "did:example:123#z6MksHh7qHWvybLg5QTPPdG2DgEjjduBDArV9EF9mRiRzMBN",
                    "proofPurpose": "assertionMethod",
                    "jws": "eyJhbGciOiJFZERTQSIsImI2NCI6ZmFsc2UsImNyaXQiOlsiYjY0Il19..YtqjEYnFENT7fNW-COD0HAACxeuQxPKAmp4nIl8jYAu__6IH2FpSxv81w-l5PvE1og50tS9tH8WyXMlXyo45CA",
                }
            },
        ],
        "proof": {
            "verificationMethod": "did:example:123#z6MksHh7qHWvybLg5QTPPdG2DgEjjduBDArV9EF9mRiRzMBN",
            "proofPurpose": "assertionMethod",
            "created": "2020-04-02T18:48:36Z",
            "domain": "example.com",
            "challenge": "d436f0c8-fbd9-4e48-bbb2-55fc5d0920a8",
            "type": "Ed25519Signature2018",
            "jws": "eyJhbGciOiAiRWREU0EiLCAiYjY0IjogZmFsc2UsICJjcml0IjogWyJiNjQiXX0..a6dB9OAI9HWc1lDoWzd1---XF_QdArVMu99N2OKnOFT2Ize8MiuVvbJCIkYHpjn3arPle-o0iMlUx3q08ES_Bg",
        },
    },
    "verkey": "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx",
}
TEST_VERIFY_OBJS = [
    TEST_VERIFY_OBJ0,
    TEST_VERIFY_OBJ1,
    TEST_VERIFY_OBJ2,
]
TEST_VERIFY_ERROR = {
    "verkey": "5yKdnU7ToTjAoRNDzfuzVTfWBH38qyhE1b9xh4v8JaWF",
    "doc": {
        "@context": [
            "https://www.w3.org/2018/credentials/v1",
            "https://www.w3.org/2018/credentials/examples/v1",
        ],
        "id": "http://example.gov/credentials/3732",
        "type": ["VerifiableCredential", "UniversityDegreeCredential"],
        "issuer": ("did:key:z6MkjRagNiMu91DduvCvgEsqLZDVzrJzFrwahc4tXLt9DoHd"),
        "issuanceDate": "2020-03-10T04:24:12.164Z",
        "credentialSubject": {
            "id": ("did:key:z6MkjRagNiMu91DduvCvgEsqLZDVzrJzFrwahc4tXLt9DoHd"),
            "degree": {
                "type": "BachelorDegree",
                "name": "Bachelor of Science and Arts",
            },
        },
        "proof": {
            "type": "Ed25519Signature2018",
            "created": "2020-04-10T21:35:35Z",
            "verificationMethod": (
                "did:key:"
                "z6MkjRagNiMu91DduvCvgEsqLZDVzrJzFrwahc"
                "4tXLt9DoHd#z6MkjRagNiMu91DduvCvgEsqLZD"
                "VzrJzFrwahc4tXLt9DoHd"
            ),
            "proofPurpose": "assertionMethod",
            "jws": (
                "eyJhbGciOiJ0RWRUZXN0RWQiLCJiNjQiOmZhbHNlLCJjcml0IjpbImI2NCJdfQ..l9d0YHjcFAH2H4dB9xlWFZQLUp"
                "ixVCWJk0eOt4CXQe1NXKWZwmhmn9OQp6YxX0a2Lffe"
                "gtYESTCJEoGVXLqWAA"
            ),
        },
    },
}
TEST_EURO_HEALTH = {
    "@context": {
        "@version": 1.1,
        "@protected": True,
        "name": "http://schema.org/name",
        "description": "http://schema.org/description",
        "EuropeanHealthInsuranceCard": {
            "@id": "https://essif-lab.pages.grnet.gr/interoperability/eidas-generic-use-case#EuropeanHealthInsuranceCard",
            "@context": {
                "@version": 1.1,
                "@protected": True,
                "id": "@id",
                "type": "@type",
                "ehic": "https://essif-lab.pages.grnet.gr/interoperability/eidas-generic-use-case#",
                "description": "http://schema.org/description",
                "name": "http://schema.org/name",
                "institutionID": "ehic:identificationNumberOfTheInstitution",
                "cardNo": "ehic:identificationNumberOfTheCard",
                "personalID": "ehic:personalIdentificationNumber",
            },
        },
        "EuropeanHealthInsuranceHolder": {
            "@id": "https://essif-lab.pages.grnet.gr/interoperability/eidas-generic-use-case#EuropeanHealthInsuranceHolder",
            "@context": {
                "@version": 1.1,
                "@protected": True,
                "schema": "http://schema.org/",
                "xsd": "http://www.w3.org/2001/XMLSchema#",
                "id": "@id",
                "type": "@type",
                "birthDate": {"@id": "schema:birthDate", "@type": "xsd:dateTime"},
                "familyName": "schema:familyName",
                "givenName": "schema:givenName",
            },
        },
        "Person": "http://schema.org/Person",
    }
}



"""Test mediate request message."""

from unittest import TestCase

from ...message_types import MEDIATE_REQUEST
from ..mediate_request import MediationRequest, MediationRequestSchema
from . import MessageTest


class TestMediateRequest(MessageTest, TestCase):
    """Test mediate request message."""

    TYPE = MEDIATE_REQUEST
    CLASS = MediationRequest
    SCHEMA = MediationRequestSchema
    VALUES = {}



import pytest
import pytest_asyncio

from acapy_agent.tests import mock

from ......messaging.request_context import RequestContext
from ......messaging.responder import MockResponder
from ......transport.inbound.receipt import MessageReceipt
from ......utils.testing import create_test_profile
from ......wallet.did_method import DIDMethods
from ...manager import DIDXManagerError
from ...messages.complete import DIDXComplete
from ...messages.problem_report import ProblemReportReason
from .. import complete_handler as test_module


@pytest_asyncio.fixture
async def request_context():
    ctx = RequestContext.test_context(await create_test_profile())
    ctx.injector.bind_instance(DIDMethods, DIDMethods())
    ctx.message_receipt = MessageReceipt()
    yield ctx


class TestDIDXCompleteHandler:
    """Class unit testing complete handler."""

    @pytest.mark.asyncio
    @mock.patch.object(test_module, "DIDXManager")
    async def test_called(self, mock_conn_mgr, request_context):
        mock_conn_mgr.return_value.accept_complete = mock.CoroutineMock()
        request_context.message = DIDXComplete()
        handler_inst = test_module.DIDXCompleteHandler()
        await handler_inst.handle(request_context, None)

        mock_conn_mgr.return_value.accept_complete.assert_called_once_with(
            request_context.message, request_context.message_receipt
        )

    @pytest.mark.asyncio
    @mock.patch.object(test_module, "DIDXManager")
    async def test_x(
        self, mock_conn_mgr, request_context, caplog: pytest.LogCaptureFixture
    ):
        mock_conn_mgr.return_value.accept_complete = mock.CoroutineMock(
            side_effect=DIDXManagerError(
                error_code=ProblemReportReason.COMPLETE_NOT_ACCEPTED.value
            )
        )
        mock_conn_mgr.return_value._logger = mock.MagicMock(exception=mock.MagicMock())
        request_context.message = DIDXComplete()
        handler_inst = test_module.DIDXCompleteHandler()
        responder = MockResponder()
        with caplog.at_level("ERROR"):
            await handler_inst.handle(request_context, responder)
        assert "Error receiving connection complete" in caplog.text



"""Credential problem report message handler."""

from .....messaging.base_handler import BaseHandler, HandlerException
from .....messaging.request_context import RequestContext
from .....messaging.responder import BaseResponder
from .....storage.error import StorageError, StorageNotFoundError
from ..manager import CredentialManager
from ..messages.credential_problem_report import CredentialProblemReport


class CredentialProblemReportHandler(BaseHandler):
    """Message handler class for problem reports."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Message handler logic for problem reports.

        Args:
            context: request context
            responder: responder callback
        """
        self._logger.debug(
            "Issue-credential v1.0 problem report handler called with context %s",
            context,
        )
        assert isinstance(context.message, CredentialProblemReport)

        # If connection is present it must be ready for use
        if context.connection_record and not context.connection_ready:
            raise HandlerException(
                "Connection used for credential problem report not ready"
            )
        elif not context.connection_record:
            raise HandlerException(
                "Connectionless not supported for credential problem report"
            )

        credential_manager = CredentialManager(context.profile)
        try:
            await credential_manager.receive_problem_report(
                context.message,
                context.connection_record.connection_id,
            )
        except (StorageError, StorageNotFoundError):
            self._logger.exception(
                "Error processing issue-credential v1.0 problem report message"
            )



"""Credential issue message handler."""

from .....core.oob_processor import OobMessageProcessor
from .....indy.holder import IndyHolderError
from .....messaging.base_handler import BaseHandler, HandlerException
from .....messaging.models.base import BaseModelError
from .....messaging.request_context import RequestContext
from .....messaging.responder import BaseResponder
from .....storage.error import StorageError
from .....utils.tracing import get_timer, trace_event
from .. import problem_report_for_record
from ..manager import CredentialManager, CredentialManagerError
from ..messages.credential_issue import CredentialIssue
from ..messages.credential_problem_report import ProblemReportReason


class CredentialIssueHandler(BaseHandler):
    """Message handler class for credential offers."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Message handler logic for credential offers.

        Args:
            context: request context
            responder: responder callback

        """
        r_time = get_timer()
        profile = context.profile
        self._logger.debug("CredentialHandler called with context %s", context)
        assert isinstance(context.message, CredentialIssue)
        self._logger.debug(
            "Received credential message: %s", context.message.serialize(as_string=True)
        )

        # If connection is present it must be ready for use
        if context.connection_record and not context.connection_ready:
            raise HandlerException("Connection used for credential not ready")

        # Find associated oob record
        oob_processor = context.inject(OobMessageProcessor)
        oob_record = await oob_processor.find_oob_record_for_inbound_message(context)

        # Either connection or oob context must be present
        if not context.connection_record and not oob_record:
            raise HandlerException(
                "No connection or associated connectionless exchange found for credential"
            )

        credential_manager = CredentialManager(profile)
        cred_ex_record = await credential_manager.receive_credential(
            context.message,
            (
                context.connection_record.connection_id
                if context.connection_record
                else None
            ),
        )  # mgr only finds, saves record: on exception, saving state null is hopeless

        r_time = trace_event(
            context.settings,
            context.message,
            outcome="CredentialIssueHandler.handle.END",
            perf_counter=r_time,
        )

        # Automatically move to next state if flag is set
        if cred_ex_record and context.settings.get("debug.auto_store_credential"):
            try:
                cred_ex_record = await credential_manager.store_credential(cred_ex_record)
            except (
                BaseModelError,
                CredentialManagerError,
                IndyHolderError,
                StorageError,
            ) as err:
                # treat failure to store as mangled on receipt hence protocol error
                self._logger.exception("Error storing issued credential")
                if cred_ex_record:
                    async with profile.session() as session:
                        await cred_ex_record.save_error_state(
                            session,
                            reason=err.roll_up,  # us: be specific
                        )
                    await responder.send_reply(
                        problem_report_for_record(
                            cred_ex_record,
                            ProblemReportReason.ISSUANCE_ABANDONED.value,  # them: vague
                        )
                    )

            (_, credential_ack_message) = await credential_manager.send_credential_ack(
                cred_ex_record
            )

            trace_event(
                context.settings,
                credential_ack_message,
                outcome="CredentialIssueHandler.handle.STORE",
                perf_counter=r_time,
            )



"""Handler for incoming forward messages."""

import json

from .....connections.base_manager import BaseConnectionManager
from .....messaging.base_handler import (
    BaseHandler,
    BaseResponder,
    HandlerException,
    RequestContext,
)
from ..manager import RoutingManager, RoutingManagerError
from ..messages.forward import Forward


class ForwardHandler(BaseHandler):
    """Handler for incoming forward messages."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Message handler implementation."""
        self._logger.debug("ForwardHandler called with context %s", context)
        assert isinstance(context.message, Forward)

        if not context.message_receipt.recipient_verkey:
            raise HandlerException("Cannot forward message: unknown recipient")
        self._logger.debug(
            "Received forward for: %s", context.message_receipt.recipient_verkey
        )

        packed = context.message.msg
        packed = json.dumps(packed).encode("ascii")
        rt_mgr = RoutingManager(context.profile)
        target = context.message.to

        try:
            recipient = await rt_mgr.get_recipient(target)
        except RoutingManagerError:
            self._logger.exception("Error resolving recipient for forwarded message")
            return

        # load connection
        connection_mgr = BaseConnectionManager(context.profile)
        connection_targets = await connection_mgr.get_connection_targets(
            connection_id=recipient.connection_id
        )
        # TODO: validate that there is 1 target, with 1 verkey. warn otherwise
        connection_verkey = connection_targets[0].recipient_keys[0]

        # Note: not currently vetting the state of the connection here
        self._logger.info(f"Forwarding message to connection: {recipient.connection_id}")

        send_status = await responder.send(
            packed,
            connection_id=recipient.connection_id,
            target_list=connection_targets,
            reply_to_verkey=connection_verkey,
        )

        # emit event that a forward message is received (may trigger webhook event)
        await context.profile.notify(
            "acapy::forward::received",
            {
                "connection_id": recipient.connection_id,
                "status": send_status.value,
                "recipient_key": context.message.to,
            },
        )



from unittest import IsolatedAsyncioTestCase

from acapy_agent.ledger.base import BaseLedger
from acapy_agent.tests import mock
from acapy_agent.wallet.askar import AskarWallet
from acapy_agent.wallet.did_info import DIDInfo
from acapy_agent.wallet.did_method import INDY, SOV, WEB
from acapy_agent.wallet.error import WalletError
from acapy_agent.wallet.key_type import ED25519

from ...utils.testing import create_test_profile


class TestAskar(IsolatedAsyncioTestCase):
    async def asyncSetUp(self) -> None:
        self.profile = await create_test_profile()

    async def test_set_did_endpoint(self):
        wallet = AskarWallet(self.profile.session)
        wallet.replace_local_did_metadata = mock.CoroutineMock()

        # Set endpoint for a Sov DID
        sov_did_info = DIDInfo("example123", "verkey", {}, SOV, ED25519.key_type)
        wallet.get_local_did = mock.CoroutineMock(return_value=sov_did_info)
        wallet.get_public_did = mock.CoroutineMock(return_value=sov_did_info)
        await wallet.set_did_endpoint(
            "did:example:123",
            "http://example.com",
            mock.MagicMock(BaseLedger, autospec=True),
        )

        # Set endpoint for an Indy DID
        indy_did_info = DIDInfo("did:indy:example", "verkey", {}, INDY, ED25519.key_type)
        wallet.get_local_did = mock.CoroutineMock(return_value=indy_did_info)
        wallet.get_public_did = mock.CoroutineMock(return_value=indy_did_info)
        await wallet.set_did_endpoint(
            "did:example:123",
            "http://example.com",
            mock.MagicMock(BaseLedger, autospec=True),
        )

        # Set endpoint for a Web DID should fail
        web_did_info = DIDInfo("did:web:example:123", "verkey", {}, WEB, ED25519.key_type)
        wallet.get_local_did = mock.CoroutineMock(return_value=web_did_info)
        wallet.get_public_did = mock.CoroutineMock(return_value=web_did_info)
        with self.assertRaises(WalletError):
            await wallet.set_did_endpoint(
                "did:example:123",
                "http://example.com",
                mock.MagicMock(BaseLedger, autospec=True),
            )



"""Classes to manage credential revocation."""

import asyncio
import json
import logging
from typing import Mapping, NamedTuple, Optional, Sequence, Text, Tuple

from ..cache.base import BaseCache
from ..connections.models.conn_record import ConnRecord
from ..core.error import BaseError
from ..core.profile import Profile, ProfileSession
from ..indy.credx.issuer import CATEGORY_REV_REG
from ..indy.issuer import IndyIssuer
from ..ledger.base import BaseLedger
from ..messaging.responder import BaseResponder
from ..protocols.endorse_transaction.v1_0.manager import (
    TransactionManager,
    TransactionManagerError,
)
from ..protocols.endorse_transaction.v1_0.util import get_endorser_connection_id
from ..protocols.issue_credential.v1_0.models.credential_exchange import (
    V10CredentialExchange,
)
from ..protocols.issue_credential.v2_0.models.cred_ex_record import V20CredExRecord
from ..protocols.revocation_notification.v1_0.models.rev_notification_record import (
    RevNotificationRecord,
)
from ..storage.error import StorageError, StorageNotFoundError
from .indy import IndyRevocation
from .models.issuer_cred_rev_record import IssuerCredRevRecord
from .models.issuer_rev_reg_record import IssuerRevRegRecord
from .util import notify_pending_cleared_event, notify_revocation_published_event

LOGGER = logging.getLogger(__name__)


class RevocationManagerError(BaseError):
    """Revocation manager error."""


class RevocationNotificationInfo(NamedTuple):
    """Revocation notification information."""

    rev_reg_id: str
    cred_rev_id: str
    thread_id: Optional[str]
    connection_id: Optional[str]
    comment: Optional[str]
    notify_version: Optional[str]


class RevocationManager:
    """Class for managing revocation operations."""

    def __init__(self, profile: Profile):
        """Initialize a RevocationManager.

        Args:
            profile: The profile instance for this revocation manager
        """
        self._profile = profile
        self._logger = logging.getLogger(__name__)

    async def revoke_credential_by_cred_ex_id(
        self,
        cred_ex_id: str,
        publish: bool = False,
        notify: bool = False,
        notify_version: Optional[str] = None,
        thread_id: Optional[str] = None,
        connection_id: Optional[str] = None,
        endorser_conn_id: Optional[str] = None,
        comment: Optional[str] = None,
        write_ledger: bool = True,
    ):
        """Revoke a credential by its credential exchange identifier at issue.

        Optionally, publish the corresponding revocation registry delta to the ledger.

        Args:
            cred_ex_id (str): The credential exchange identifier.
            publish (bool, optional): Whether to publish the resulting revocation
                registry delta, along with any revocations pending against it.
                Defaults to False.
            notify (bool, optional): Whether to notify the affected parties about the
                revocation. Defaults to False.
            notify_version (str, optional): The version of the notification to use.
                Defaults to None.
            thread_id (str, optional): The thread identifier for the revocation process.
                Defaults to None.
            connection_id (str, optional): The connection identifier for the revocation
                process. Defaults to None.
            endorser_conn_id (str, optional): The endorser connection identifier for the
                revocation process. Defaults to None.
            comment (str, optional): Additional comment or reason for the revocation.
                Defaults to None.
            write_ledger (bool, optional): Whether to write the revocation to the ledger.
                Defaults to True.

        Raises:
            RevocationManagerError: If no issuer credential revocation record is found
                for the given credential exchange identifier.

        Returns:
            The result of the `revoke_credential` method.

        """
        try:
            async with self._profile.session() as session:
                rec = await IssuerCredRevRecord.retrieve_by_cred_ex_id(
                    session,
                    cred_ex_id,
                )
        except StorageNotFoundError as err:
            raise RevocationManagerError(
                "No issuer credential revocation record found for "
                f"credential exchange id {cred_ex_id}"
            ) from err

        return await self.revoke_credential(
            rev_reg_id=rec.rev_reg_id,
            cred_rev_id=rec.cred_rev_id,
            publish=publish,
            notify=notify,
            notify_version=notify_version,
            thread_id=thread_id,
            connection_id=connection_id,
            endorser_conn_id=endorser_conn_id,
            comment=comment,
            write_ledger=write_ledger,
        )

    async def _prepare_revocation_notification(
        self,
        revoc_notif_info: RevocationNotificationInfo,
    ):
        """Saves the revocation notification record, and thread_id if not provided."""
        thread_id = (
            revoc_notif_info.thread_id
            or f"indy::{revoc_notif_info.rev_reg_id}::{revoc_notif_info.cred_rev_id}"
        )
        rev_notify_rec = RevNotificationRecord(
            rev_reg_id=revoc_notif_info.rev_reg_id,
            cred_rev_id=revoc_notif_info.cred_rev_id,
            thread_id=thread_id,
            connection_id=revoc_notif_info.connection_id,
            comment=revoc_notif_info.comment,
            version=revoc_notif_info.notify_version,
        )
        async with self._profile.session() as session:
            await rev_notify_rec.save(session, reason="New revocation notification")

    async def _get_endorsement_txn_for_revocation(
        self, endorser_conn_id: str, issuer_rr_upd: IssuerRevRegRecord
    ):
        async with self._profile.session() as session:
            try:
                connection_record = await ConnRecord.retrieve_by_id(
                    session, endorser_conn_id
                )
            except StorageNotFoundError:
                raise RevocationManagerError(
                    f"No endorser connection record found for id: {endorser_conn_id}"
                )
            endorser_info = await connection_record.metadata_get(session, "endorser_info")
        endorser_did = endorser_info["endorser_did"]
        return await issuer_rr_upd.send_entry(
            self._profile,
            write_ledger=False,
            endorser_did=endorser_did,
        )

    async def revoke_credential(
        self,
        rev_reg_id: str,
        cred_rev_id: str,
        publish: bool = False,
        notify: bool = False,
        notify_version: Optional[str] = None,
        thread_id: Optional[str] = None,
        connection_id: Optional[str] = None,
        endorser_conn_id: Optional[str] = None,
        comment: Optional[str] = None,
        write_ledger: bool = True,
    ) -> Optional[dict]:
        """Revoke a credential.

        Optionally, publish the corresponding revocation registry delta to the ledger.

        Args:
            rev_reg_id (str): The revocation registry id.
            cred_rev_id (str): The credential revocation id.
            publish (bool, optional): Whether to publish the resulting revocation
                registry delta, along with any revocations pending against it.
                Defaults to False.
            notify (bool, optional): Whether to send a revocation notification.
                Defaults to False.
            notify_version (str, optional): The version of the revocation notification.
                Defaults to None.
            thread_id (str, optional): The thread id for the revocation notification.
                Defaults to None.
            connection_id (str, optional): The connection id for the revocation
                notification. Defaults to None.
            endorser_conn_id (str, optional): The endorser connection id.
                Defaults to None.
            comment (str, optional): Additional comment for the revocation notification.
                Defaults to None.
            write_ledger (bool, optional): Whether to write the revocation entry to the
                ledger. Defaults to True.

        Returns:
            Optional[dict]: The revocation entry response if publish is True and
                write_ledger is True, otherwise None.
        """
        issuer = self._profile.inject(IndyIssuer)
        revoc = IndyRevocation(self._profile)

        issuer_rr_rec = await revoc.get_issuer_rev_reg_record(rev_reg_id)
        if not issuer_rr_rec:
            raise RevocationManagerError(
                f"No revocation registry record found for id: {rev_reg_id}"
            )

        if notify:
            await self._prepare_revocation_notification(
                RevocationNotificationInfo(
                    rev_reg_id=rev_reg_id,
                    cred_rev_id=cred_rev_id,
                    thread_id=thread_id,
                    connection_id=connection_id,
                    comment=comment,
                    notify_version=notify_version,
                ),
            )

        if not publish:
            # If not publishing, just mark the revocation as pending.
            async with self._profile.transaction() as txn:
                await issuer_rr_rec.mark_pending(txn, cred_rev_id)
                await txn.commit()
            return None

        rev_reg = await revoc.get_ledger_registry(rev_reg_id)
        await rev_reg.get_or_fetch_local_tails_path()
        # pick up pending revocations on input revocation registry
        crids = (issuer_rr_rec.pending_pub or []) + [cred_rev_id]
        (delta_json, _) = await issuer.revoke_credentials(
            issuer_rr_rec.cred_def_id,
            issuer_rr_rec.revoc_reg_id,
            issuer_rr_rec.tails_local_path,
            crids,
        )

        # Update the revocation registry record with the new delta
        # and clear pending revocations
        async with self._profile.transaction() as txn:
            issuer_rr_upd = await IssuerRevRegRecord.retrieve_by_id(
                txn, issuer_rr_rec.record_id, for_update=True
            )
            if delta_json:
                issuer_rr_upd.revoc_reg_entry = json.loads(delta_json)
            await issuer_rr_upd.clear_pending(txn, crids)
            await txn.commit()

        await self.set_cred_revoked_state(rev_reg_id, crids)

        # Revocation list needs to be updated on ledger
        if delta_json:
            # Can write to ledger directly
            if write_ledger:
                rev_entry_resp = await issuer_rr_upd.send_entry(self._profile)
                await notify_revocation_published_event(
                    self._profile, rev_reg_id, [cred_rev_id]
                )
                return rev_entry_resp
            # Author --> Need endorsed transaction for revocation
            else:
                return await self._get_endorsement_txn_for_revocation(
                    endorser_conn_id, issuer_rr_upd
                )

    async def update_rev_reg_revoked_state(
        self,
        apply_ledger_update: bool,
        rev_reg_record: IssuerRevRegRecord,
        genesis_transactions: dict,
    ) -> Tuple[dict, dict, dict]:
        """Request handler to fix ledger entry of credentials revoked against registry.

        Args:
            apply_ledger_update (bool): Whether to apply an update to the ledger.
            rev_reg_record (IssuerRevRegRecord): The revocation registry record.
            genesis_transactions (dict): The genesis transactions.

        Returns:
            Tuple[dict, dict, dict]: A tuple containing the number of credentials posted
                to the ledger.

        """
        return await rev_reg_record.fix_ledger_entry(
            self._profile,
            apply_ledger_update,
            genesis_transactions,
        )

    async def publish_pending_revocations(
        self,
        rrid2crid: Mapping[Text, Sequence[Text]] = None,
        write_ledger: bool = True,
        connection_id: Optional[str] = None,
    ) -> Tuple[Optional[dict], Mapping[Text, Sequence[Text]]]:
        """Publish pending revocations to the ledger.

        Args:
            rrid2crid: Mapping from revocation registry identifiers to all credential
                revocation identifiers within each to publish. Specify null/empty map
                for all revocation registries. Specify empty sequence per revocation
                registry identifier for all pending within the revocation registry;
                e.g.,

                {} - publish all pending revocations from all revocation registries

                {
                    "R17v42T4pk...:4:R17v42T4pk...:3:CL:19:tag:CL_ACCUM:0": [],
                    "R17v42T4pk...:4:R17v42T4pk...:3:CL:19:tag:CL_ACCUM:1": ["1", "2"]

                } - publish:
                    - all pending revocations from all revocation registry tagged 0
                    - pending ["1", "2"] from revocation registry tagged 1
                    - no pending revocations from any other revocation registries.
            write_ledger: whether to write the revocation registry entry to the ledger
            connection_id: connection identifier for endorser connection to use

        Returns: mapping from each revocation registry id to its cred rev ids published.
        """
        result = {}
        issuer = self._profile.inject(IndyIssuer)
        rev_entry_responses = []
        async with self._profile.session() as session:
            issuer_rr_recs = await IssuerRevRegRecord.query_by_pending(session)

        for issuer_rr_rec in issuer_rr_recs:
            rrid = issuer_rr_rec.revoc_reg_id
            if rrid2crid:
                if rrid not in rrid2crid:
                    continue
                limit_crids = rrid2crid[rrid]
            else:
                limit_crids = ()
            crids = set(issuer_rr_rec.pending_pub or ())
            if limit_crids:
                crids = crids.intersection(limit_crids)
            if crids:
                crids = list(crids)
                (delta_json, failed_crids) = await issuer.revoke_credentials(
                    issuer_rr_rec.cred_def_id,
                    issuer_rr_rec.revoc_reg_id,
                    issuer_rr_rec.tails_local_path,
                    crids,
                )
                async with self._profile.transaction() as txn:
                    issuer_rr_upd = await IssuerRevRegRecord.retrieve_by_id(
                        txn, issuer_rr_rec.record_id, for_update=True
                    )
                    if delta_json:
                        issuer_rr_upd.revoc_reg_entry = json.loads(delta_json)
                    await issuer_rr_upd.clear_pending(txn, crids)
                    await txn.commit()
                await self.set_cred_revoked_state(issuer_rr_rec.revoc_reg_id, crids)
                if delta_json:
                    if connection_id:
                        async with self._profile.session() as session:
                            try:
                                connection_record = await ConnRecord.retrieve_by_id(
                                    session, connection_id
                                )
                            except StorageNotFoundError:
                                raise RevocationManagerError(
                                    "No endorser connection record found "
                                    f"for id: {connection_id}"
                                )
                            endorser_info = await connection_record.metadata_get(
                                session, "endorser_info"
                            )
                        endorser_did = endorser_info["endorser_did"]
                        rev_entry_responses.append(
                            await issuer_rr_upd.send_entry(
                                self._profile,
                                write_ledger=write_ledger,
                                endorser_did=endorser_did,
                            )
                        )
                    else:
                        rev_entry_responses.append(
                            await issuer_rr_upd.send_entry(self._profile)
                        )
                        await notify_revocation_published_event(
                            self._profile, issuer_rr_rec.revoc_reg_id, crids
                        )
                published = sorted(crid for crid in crids if crid not in failed_crids)
                result[issuer_rr_rec.revoc_reg_id] = published

        return rev_entry_responses, result

    async def clear_pending_revocations(
        self, purge: Mapping[Text, Sequence[Text]] = None
    ) -> Mapping[Text, Sequence[Text]]:
        """Clear pending revocation publications.

        Args:
            purge: Mapping from revocation registry identifiers to all credential
                revocation identifiers within each to clear. Specify null/empty map
                for all revocation registries. Specify empty sequence per revocation
                registry identifier for all pending within the revocation registry;
                e.g.,

                {} - clear all pending revocations from all revocation registries

                {
                    "R17v42T4pk...:4:R17v42T4pk...:3:CL:19:tag:CL_ACCUM:0": [],
                    "R17v42T4pk...:4:R17v42T4pk...:3:CL:19:tag:CL_ACCUM:1": ["1", "2"]

                } - clear
                    - all pending revocations from all revocation registry tagged 0
                    - pending ["1", "2"] from revocation registry tagged 1
                    - no pending revocations from any other revocation registries.

        Returns:
            mapping from revocation registry id to its remaining
            cred rev ids still marked pending, omitting revocation registries
            with no remaining pending publications.

        """
        result = {}
        notify = []

        async with self._profile.transaction() as txn:
            issuer_rr_recs = await IssuerRevRegRecord.query_by_pending(txn)
            for issuer_rr_rec in issuer_rr_recs:
                if purge and issuer_rr_rec.revoc_reg_id not in purge:
                    continue
                rrid = issuer_rr_rec.revoc_reg_id
                await issuer_rr_rec.clear_pending(txn, (purge or {}).get(rrid))
                result[rrid] = issuer_rr_rec.pending_pub
                notify.append(rrid)
            await txn.commit()

        for rrid in notify:
            await notify_pending_cleared_event(self._profile, rrid)

        return result

    async def set_cred_revoked_state(
        self, rev_reg_id: str, cred_rev_ids: Sequence[str]
    ) -> None:
        """Update credentials state to credential_revoked.

        Args:
            rev_reg_id: revocation registry ID
            cred_rev_ids: list of credential revocation IDs

        Returns:
            None

        """
        for cred_rev_id in cred_rev_ids:
            cred_ex_id = None

            try:
                async with self._profile.transaction() as txn:
                    rev_rec = await IssuerCredRevRecord.retrieve_by_ids(
                        txn, rev_reg_id, cred_rev_id, for_update=True
                    )
                    cred_ex_id = rev_rec.cred_ex_id
                    cred_ex_version = rev_rec.cred_ex_version
                    rev_rec.state = IssuerCredRevRecord.STATE_REVOKED
                    await rev_rec.save(txn, reason="revoke credential")
                    await txn.commit()
            except StorageNotFoundError:
                continue

            async with self._profile.transaction() as txn:
                if (
                    not cred_ex_version
                    or cred_ex_version == IssuerCredRevRecord.VERSION_1
                ):
                    try:
                        cred_ex_record = await V10CredentialExchange.retrieve_by_id(
                            txn, cred_ex_id, for_update=True
                        )
                        cred_ex_record.state = (
                            V10CredentialExchange.STATE_CREDENTIAL_REVOKED
                        )
                        await cred_ex_record.save(txn, reason="revoke credential")
                        await txn.commit()
                        continue  # skip 2.0 record check
                    except StorageNotFoundError:
                        pass

                if (
                    not cred_ex_version
                    or cred_ex_version == IssuerCredRevRecord.VERSION_2
                ):
                    try:
                        cred_ex_record = await V20CredExRecord.retrieve_by_id(
                            txn, cred_ex_id, for_update=True
                        )
                        cred_ex_record.state = V20CredExRecord.STATE_CREDENTIAL_REVOKED
                        await cred_ex_record.save(txn, reason="revoke credential")
                        await txn.commit()
                    except StorageNotFoundError:
                        pass

    async def _get_endorser_info(self) -> Tuple[Optional[str], Optional[ConnRecord]]:
        connection_id = await get_endorser_connection_id(self._profile)

        endorser_did = None
        async with self._profile.session() as session:
            connection_record = await ConnRecord.retrieve_by_id(session, connection_id)
            endorser_info = await connection_record.metadata_get(session, "endorser_info")
        endorser_did = endorser_info.get("endorser_did")

        return endorser_did, connection_record

    async def fix_and_publish_from_invalid_accum_err(self, err_msg: str):
        """Fix and publish revocation registry entries from invalid accumulator error."""
        cache = self._profile.inject_or(BaseCache)

        async def check_retry(accum):
            """Used to manage retries for fixing revocation registry entries."""
            retry_value = await cache.get(accum)
            if not retry_value:
                await cache.set(accum, 5)
            else:
                if retry_value > 0:
                    await cache.set(accum, retry_value - 1)
                else:
                    LOGGER.error(
                        f"Revocation registry entry transaction failed for {accum}"
                    )

        def get_genesis_transactions():
            """Get the genesis transactions needed for fixing broken accum."""
            genesis_transactions = self._profile.context.settings.get(
                "ledger.genesis_transactions"
            )
            if not genesis_transactions:
                write_ledger = self._profile.context.injector.inject(BaseLedger)
                pool = write_ledger.pool
                genesis_transactions = pool.genesis_txns
            return genesis_transactions

        async def sync_accumulator(session: ProfileSession):
            """Sync the local accumulator with the ledger and create recovery txn."""
            rev_reg_record = await IssuerRevRegRecord.retrieve_by_id(
                session, rev_reg_entry.name
            )

            # Fix and get the recovery transaction
            (
                rev_reg_delta,
                recovery_txn,
                applied_txn,
            ) = await rev_reg_record.fix_ledger_entry(
                self._profile, False, genesis_transactions
            )

            # Update locally assuming ledger write will succeed
            rev_reg = await session.handle.fetch(
                CATEGORY_REV_REG,
                rev_reg_entry.value_json["revoc_reg_id"],
                for_update=True,
            )
            new_value_json = rev_reg.value_json
            new_value_json["value"]["accum"] = recovery_txn["value"]["accum"]
            await session.handle.replace(
                CATEGORY_REV_REG,
                rev_reg.name,
                json.dumps(new_value_json),
                rev_reg.tags,
            )

            return rev_reg_record, recovery_txn

        async def create_and_send_endorser_txn():
            """Create and send the endorser transaction again."""
            async with ledger:
                # Create the revocation registry entry
                rev_entry_res = await ledger.send_revoc_reg_entry(
                    rev_reg_entry.value_json["revoc_reg_id"],
                    "CL_ACCUM",
                    recovery_txn,
                    rev_reg_record.issuer_did,
                    write_ledger=False,
                    endorser_did=endorser_did,
                )

            # Send the transaction to the endorser again with recovery txn
            transaction_manager = TransactionManager(self._profile)
            try:
                revo_transaction = await transaction_manager.create_record(
                    messages_attach=rev_entry_res["result"],
                    connection_id=connection.connection_id,
                )
                (
                    revo_transaction,
                    revo_transaction_request,
                ) = await transaction_manager.create_request(transaction=revo_transaction)
            except (StorageError, TransactionManagerError) as err:
                raise RevocationManagerError(err.roll_up) from err

            responder = self._profile.inject_or(BaseResponder)
            if not responder:
                raise RevocationManagerError(
                    "No responder found. Unable to send transaction request"
                )
            await responder.send(
                revo_transaction_request,
                connection_id=connection.connection_id,
            )

        async with self._profile.session() as session:
            rev_reg_records = await session.handle.fetch_all(
                IssuerRevRegRecord.RECORD_TYPE
            )
            # Cycle through all rev_rev_def records to find the offending accumulator
            for rev_reg_entry in rev_reg_records:
                ledger = session.inject_or(BaseLedger)
                # Get the value from the ledger
                async with ledger:
                    (accum_response, _) = await ledger.get_revoc_reg_delta(
                        rev_reg_entry.value_json["revoc_reg_id"]
                    )
                    accum = accum_response.get("value", {}).get("accum")

                # If the accum from the ledger matches the error message, fix it
                if accum and accum in err_msg:
                    await check_retry(accum)

                    # Get the genesis transactions needed for fix
                    genesis_transactions = get_genesis_transactions()

                    # We know this needs endorsement
                    endorser_did, connection = await self._get_endorser_info()
                    rev_reg_record, recovery_txn = await sync_accumulator(session=session)
                    await create_and_send_endorser_txn()

                    # Some time in between re-tries
                    await asyncio.sleep(1)



from unittest import IsolatedAsyncioTestCase

from acapy_agent.tests import mock

from ...askar.profile import AskarProfile
from ...messaging.responder import BaseResponder
from ...utils.testing import create_test_profile
from ...wallet.models.wallet_record import WalletRecord
from ..manager import MultitenantManager


class TestMultitenantManager(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.profile = await create_test_profile()
        self.context = self.profile.context

        self.responder = mock.CoroutineMock(send=mock.CoroutineMock())
        self.context.injector.bind_instance(BaseResponder, self.responder)

        self.manager = MultitenantManager(self.profile)

    async def test_get_wallet_profile_returns_from_cache(self):
        wallet_record = WalletRecord(wallet_id="test")
        self.manager._profiles.put("test", (await create_test_profile()))

        with mock.patch("acapy_agent.config.wallet.wallet_config") as wallet_config:
            profile = await self.manager.get_wallet_profile(
                self.profile.context, wallet_record
            )
            assert profile is self.manager._profiles.get("test")
            wallet_config.assert_not_called()

    async def test_get_wallet_profile_not_in_cache(self):
        wallet_record = WalletRecord(wallet_id="test", settings={})
        self.manager._profiles.put("test", (await create_test_profile()))
        self.profile.context.update_settings(
            {"admin.webhook_urls": ["http://localhost:8020"]}
        )

        with mock.patch("acapy_agent.config.wallet.wallet_config") as wallet_config:
            profile = await self.manager.get_wallet_profile(
                self.profile.context, wallet_record
            )
            assert profile is self.manager._profiles.get("test")
            wallet_config.assert_not_called()

    async def test_get_wallet_profile_settings(self):
        extra_settings = {"extra_settings": "extra_settings"}
        all_wallet_record_settings = [
            {
                "wallet_record_settings": "wallet_record_settings",
                "wallet.dispatch_type": "default",
            },
            {
                "wallet_record_settings": "wallet_record_settings",
                "wallet.dispatch_type": "default",
                "wallet.webhook_urls": ["https://localhost:8090"],
            },
            {
                "wallet_record_settings": "wallet_record_settings",
                "wallet.dispatch_type": "both",
            },
            {
                "wallet_record_settings": "wallet_record_settings",
                "wallet.dispatch_type": "both",
                "wallet.webhook_urls": ["https://localhost:8090"],
            },
        ]

        async def side_effect(context, provision):
            return (await create_test_profile(settings=None, context=context), None)

        for idx, wallet_record_settings in enumerate(all_wallet_record_settings):
            wallet_record = WalletRecord(
                wallet_id=f"test.{idx}",
                settings=wallet_record_settings,
            )

            with mock.patch(
                "acapy_agent.multitenant.manager.wallet_config"
            ) as wallet_config:
                wallet_config.side_effect = side_effect
                profile = await self.manager.get_wallet_profile(
                    self.profile.context, wallet_record, extra_settings
                )

                assert (
                    profile.settings.get("wallet_record_settings")
                    == "wallet_record_settings"
                )
                assert profile.settings.get("extra_settings") == "extra_settings"

    async def test_get_wallet_profile_settings_reset(self):
        wallet_record = WalletRecord(
            wallet_id="test",
            settings={},
        )

        with mock.patch("acapy_agent.multitenant.manager.wallet_config") as wallet_config:

            async def side_effect(context, provision):
                return (await create_test_profile(settings=None, context=context), None)

            wallet_config.side_effect = side_effect

            self.profile.context.update_settings(
                {
                    "wallet.recreate": True,
                    "wallet.seed": "test_seed",
                    "wallet.name": "test_name",
                    "wallet.type": "test_type",
                    "wallet.rekey": "test_rekey",
                    "mediation.open": True,
                    "mediation.invite": "http://invite.com",
                    "mediation.default_id": "24a96ef5",
                    "mediation.clear": True,
                }
            )

            profile = await self.manager.get_wallet_profile(
                self.profile.context, wallet_record
            )

            assert profile.settings.get("wallet.recreate") is False
            assert profile.settings.get("wallet.seed") is None
            assert profile.settings.get("wallet.rekey") is None
            assert profile.settings.get("wallet.name") is None
            assert profile.settings.get("wallet.type") is None
            assert profile.settings.get("mediation.open") is None
            assert profile.settings.get("mediation.invite") is None
            assert profile.settings.get("mediation.default_id") is None
            assert profile.settings.get("mediation.clear") is None

    async def test_get_wallet_profile_settings_reset_overwrite(self):
        wallet_record = WalletRecord(
            wallet_id="test",
            settings={
                "wallet.recreate": True,
                "wallet.seed": "test_seed",
                "wallet.name": "test_name",
                "wallet.type": "test_type",
                "wallet.rekey": "test_rekey",
                "mediation.open": True,
                "mediation.invite": "http://invite.com",
                "mediation.default_id": "24a96ef5",
                "mediation.clear": True,
            },
        )

        with mock.patch("acapy_agent.multitenant.manager.wallet_config") as wallet_config:

            async def side_effect(context, provision):
                return (await create_test_profile(settings=None, context=context), None)

            wallet_config.side_effect = side_effect

            profile = await self.manager.get_wallet_profile(
                self.profile.context, wallet_record
            )

            assert profile.settings.get("wallet.recreate") is True
            assert profile.settings.get("wallet.seed") == "test_seed"
            assert profile.settings.get("wallet.rekey") == "test_rekey"
            assert profile.settings.get("wallet.name") == "test_name"
            assert profile.settings.get("wallet.type") == "test_type"
            assert profile.settings.get("mediation.open") is True
            assert profile.settings.get("mediation.invite") == "http://invite.com"
            assert profile.settings.get("mediation.default_id") == "24a96ef5"
            assert profile.settings.get("mediation.clear") is True

    async def test_update_wallet_update_wallet_profile(self):
        with (
            mock.patch.object(WalletRecord, "retrieve_by_id") as retrieve_by_id,
            mock.patch.object(WalletRecord, "save") as wallet_record_save,
        ):
            wallet_id = "test-wallet-id"
            wallet_profile = await create_test_profile()
            self.manager._profiles.put("test-wallet-id", wallet_profile)
            retrieve_by_id.return_value = WalletRecord(
                wallet_id=wallet_id,
                settings={
                    "wallet.webhook_urls": ["test-webhook-url"],
                    "wallet.dispatch_type": "both",
                },
            )

            new_settings = {
                "wallet.webhook_urls": ["new-webhook-url"],
                "wallet.dispatch_type": "default",
            }
            wallet_record = await self.manager.update_wallet(wallet_id, new_settings)

            wallet_record_save.assert_called_once()

            assert isinstance(wallet_record, WalletRecord)
            assert wallet_record.wallet_webhook_urls == ["new-webhook-url"]
            assert wallet_record.wallet_dispatch_type == "default"
            assert wallet_profile.settings.get("wallet.webhook_urls") == [
                "new-webhook-url"
            ]
            assert wallet_profile.settings.get("wallet.dispatch_type") == "default"

    async def test_remove_wallet_profile(self):
        test_profile = await create_test_profile()
        self.manager._profiles.put(test_profile.name, test_profile)

        with mock.patch.object(AskarProfile, "remove") as profile_remove:
            await self.manager.remove_wallet_profile(test_profile)
            assert not self.manager._profiles.has(test_profile.name)
            profile_remove.assert_called_once_with()



"""Authentication decorators for the admin API."""

import functools
import re
from typing import List, Optional, Pattern

from aiohttp import web

from ...utils import general as general_utils
from ..request_context import AdminRequestContext


def admin_authentication(handler):
    """Decorator to require authentication via admin API key.

    The decorator will check for a valid x-api-key header and
    reject the request if it is missing or invalid.
    If the agent is running in insecure mode, the request will be allowed without a key.
    """

    @functools.wraps(handler)
    async def admin_auth(request):
        context: AdminRequestContext = request["context"]
        profile = context.profile
        header_admin_api_key = request.headers.get("x-api-key")
        valid_key = general_utils.const_compare(
            profile.settings.get("admin.admin_api_key"), header_admin_api_key
        )
        insecure_mode = bool(profile.settings.get("admin.admin_insecure_mode"))

        # We have to allow OPTIONS method access to paths without a key since
        # browsers performing CORS requests will never include the original
        # x-api-key header from the method that triggered the preflight
        # OPTIONS check.
        if insecure_mode or valid_key or (request.method == "OPTIONS"):
            return await handler(request)
        else:
            raise web.HTTPUnauthorized(
                reason="API Key invalid or missing",
                text="API Key invalid or missing",
            )

    return admin_auth


def tenant_authentication(handler):
    """Decorator to enable non-admin authentication.

    The decorator will:
    - check for a valid bearer token in the Autorization header if running
    in multi-tenant mode
    - check for a valid x-api-key header if running in single-tenant mode
    - check if the base wallet has access to the requested path if running
    in multi-tenant mode
    """

    @functools.wraps(handler)
    async def tenant_auth(request):
        context: AdminRequestContext = request["context"]
        profile = context.profile
        authorization_header = request.headers.get("Authorization")
        header_admin_api_key = request.headers.get("x-api-key")
        valid_key = general_utils.const_compare(
            profile.settings.get("admin.admin_api_key"), header_admin_api_key
        )
        insecure_mode = bool(profile.settings.get("admin.admin_insecure_mode"))
        multitenant_enabled = profile.settings.get("multitenant.enabled")
        base_wallet_routes = profile.settings.get("multitenant.base_wallet_routes")
        base_wallet_allowed_route = _base_wallet_route_access(
            [base_wallet_routes]
            if isinstance(base_wallet_routes, str)
            else base_wallet_routes,
            request.path,
        )

        # CORS fix: allow OPTIONS method access to paths without a token
        if (
            (multitenant_enabled and authorization_header)
            or (not multitenant_enabled and valid_key)
            or (multitenant_enabled and valid_key and base_wallet_allowed_route)
            or (insecure_mode and not multitenant_enabled)
            or request.method == "OPTIONS"
        ):
            return await handler(request)
        else:
            auth_mode = "Authorization token" if multitenant_enabled else "API key"
            raise web.HTTPUnauthorized(
                reason=f"{auth_mode} missing or invalid",
                text=f"{auth_mode} missing or invalid",
            )

    return tenant_auth


def _base_wallet_route_access(additional_routes: List[str], request_path: str) -> bool:
    """Check if request path matches additional routes."""
    additional_routes_pattern = (
        _build_additional_routes_pattern(additional_routes) if additional_routes else None
    )
    return _matches_additional_routes(additional_routes_pattern, request_path)


def _build_additional_routes_pattern(pattern_list: List[str]) -> Optional[Pattern]:
    """Build pattern from space delimited list of paths."""
    # create array and add word boundary to avoid false positives
    all_paths = []
    for pattern in pattern_list:
        paths = pattern.split(" ")
        all_paths = all_paths + paths
    return re.compile("^((?:)" + "|".join(all_paths) + ")$")


def _matches_additional_routes(pattern: Pattern, path: str) -> bool:
    """Matches request path to provided pattern."""
    if pattern and path:
        return bool(pattern.match(path))
    return False



ODRL = {
    "@context": {
        "odrl": "http://www.w3.org/ns/odrl/2/",
        "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
        "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
        "owl": "http://www.w3.org/2002/07/owl#",
        "skos": "http://www.w3.org/2004/02/skos/core#",
        "dct": "http://purl.org/dc/terms/",
        "xsd": "http://www.w3.org/2001/XMLSchema#",
        "vcard": "http://www.w3.org/2006/vcard/ns#",
        "foaf": "http://xmlns.com/foaf/0.1/",
        "schema": "http://schema.org/",
        "cc": "http://creativecommons.org/ns#",
        "uid": "@id",
        "type": "@type",
        "Policy": "odrl:Policy",
        "Rule": "odrl:Rule",
        "profile": {"@type": "@id", "@id": "odrl:profile"},
        "inheritFrom": {"@type": "@id", "@id": "odrl:inheritFrom"},
        "ConflictTerm": "odrl:ConflictTerm",
        "conflict": {"@type": "@vocab", "@id": "odrl:conflict"},
        "perm": "odrl:perm",
        "prohibit": "odrl:prohibit",
        "invalid": "odrl:invalid",
        "Agreement": "odrl:Agreement",
        "Assertion": "odrl:Assertion",
        "Offer": "odrl:Offer",
        "Privacy": "odrl:Privacy",
        "Request": "odrl:Request",
        "Set": "odrl:Set",
        "Ticket": "odrl:Ticket",
        "Asset": "odrl:Asset",
        "AssetCollection": "odrl:AssetCollection",
        "relation": {"@type": "@id", "@id": "odrl:relation"},
        "hasPolicy": {"@type": "@id", "@id": "odrl:hasPolicy"},
        "target": {"@type": "@id", "@id": "odrl:target"},
        "output": {"@type": "@id", "@id": "odrl:output"},
        "partOf": {"@type": "@id", "@id": "odrl:partOf"},
        "source": {"@type": "@id", "@id": "odrl:source"},
        "Party": "odrl:Party",
        "PartyCollection": "odrl:PartyCollection",
        "function": {"@type": "@vocab", "@id": "odrl:function"},
        "PartyScope": "odrl:PartyScope",
        "assignee": {"@type": "@id", "@id": "odrl:assignee"},
        "assigner": {"@type": "@id", "@id": "odrl:assigner"},
        "assigneeOf": {"@type": "@id", "@id": "odrl:assigneeOf"},
        "assignerOf": {"@type": "@id", "@id": "odrl:assignerOf"},
        "attributedParty": {"@type": "@id", "@id": "odrl:attributedParty"},
        "attributingParty": {"@type": "@id", "@id": "odrl:attributingParty"},
        "compensatedParty": {"@type": "@id", "@id": "odrl:compensatedParty"},
        "compensatingParty": {"@type": "@id", "@id": "odrl:compensatingParty"},
        "consentingParty": {"@type": "@id", "@id": "odrl:consentingParty"},
        "consentedParty": {"@type": "@id", "@id": "odrl:consentedParty"},
        "informedParty": {"@type": "@id", "@id": "odrl:informedParty"},
        "informingParty": {"@type": "@id", "@id": "odrl:informingParty"},
        "trackingParty": {"@type": "@id", "@id": "odrl:trackingParty"},
        "trackedParty": {"@type": "@id", "@id": "odrl:trackedParty"},
        "contractingParty": {"@type": "@id", "@id": "odrl:contractingParty"},
        "contractedParty": {"@type": "@id", "@id": "odrl:contractedParty"},
        "Action": "odrl:Action",
        "action": {"@type": "@vocab", "@id": "odrl:action"},
        "includedIn": {"@type": "@id", "@id": "odrl:includedIn"},
        "implies": {"@type": "@id", "@id": "odrl:implies"},
        "Permission": "odrl:Permission",
        "permission": {"@type": "@id", "@id": "odrl:permission"},
        "Prohibition": "odrl:Prohibition",
        "prohibition": {"@type": "@id", "@id": "odrl:prohibition"},
        "obligation": {"@type": "@id", "@id": "odrl:obligation"},
        "use": "odrl:use",
        "grantUse": "odrl:grantUse",
        "aggregate": "odrl:aggregate",
        "annotate": "odrl:annotate",
        "anonymize": "odrl:anonymize",
        "archive": "odrl:archive",
        "concurrentUse": "odrl:concurrentUse",
        "derive": "odrl:derive",
        "digitize": "odrl:digitize",
        "display": "odrl:display",
        "distribute": "odrl:distribute",
        "execute": "odrl:execute",
        "extract": "odrl:extract",
        "give": "odrl:give",
        "index": "odrl:index",
        "install": "odrl:install",
        "modify": "odrl:modify",
        "move": "odrl:move",
        "play": "odrl:play",
        "present": "odrl:present",
        "print": "odrl:print",
        "read": "odrl:read",
        "reproduce": "odrl:reproduce",
        "sell": "odrl:sell",
        "stream": "odrl:stream",
        "textToSpeech": "odrl:textToSpeech",
        "transfer": "odrl:transfer",
        "transform": "odrl:transform",
        "translate": "odrl:translate",
        "Duty": "odrl:Duty",
        "duty": {"@type": "@id", "@id": "odrl:duty"},
        "consequence": {"@type": "@id", "@id": "odrl:consequence"},
        "remedy": {"@type": "@id", "@id": "odrl:remedy"},
        "acceptTracking": "odrl:acceptTracking",
        "attribute": "odrl:attribute",
        "compensate": "odrl:compensate",
        "delete": "odrl:delete",
        "ensureExclusivity": "odrl:ensureExclusivity",
        "include": "odrl:include",
        "inform": "odrl:inform",
        "nextPolicy": "odrl:nextPolicy",
        "obtainConsent": "odrl:obtainConsent",
        "reviewPolicy": "odrl:reviewPolicy",
        "uninstall": "odrl:uninstall",
        "watermark": "odrl:watermark",
        "Constraint": "odrl:Constraint",
        "LogicalConstraint": "odrl:LogicalConstraint",
        "constraint": {"@type": "@id", "@id": "odrl:constraint"},
        "refinement": {"@type": "@id", "@id": "odrl:refinement"},
        "Operator": "odrl:Operator",
        "operator": {"@type": "@vocab", "@id": "odrl:operator"},
        "RightOperand": "odrl:RightOperand",
        "rightOperand": "odrl:rightOperand",
        "rightOperandReference": {
            "@type": "xsd:anyURI",
            "@id": "odrl:rightOperandReference",
        },
        "LeftOperand": "odrl:LeftOperand",
        "leftOperand": {"@type": "@vocab", "@id": "odrl:leftOperand"},
        "unit": "odrl:unit",
        "dataType": {"@type": "xsd:anyType", "@id": "odrl:datatype"},
        "status": "odrl:status",
        "absolutePosition": "odrl:absolutePosition",
        "absoluteSpatialPosition": "odrl:absoluteSpatialPosition",
        "absoluteTemporalPosition": "odrl:absoluteTemporalPosition",
        "absoluteSize": "odrl:absoluteSize",
        "count": "odrl:count",
        "dateTime": "odrl:dateTime",
        "delayPeriod": "odrl:delayPeriod",
        "deliveryChannel": "odrl:deliveryChannel",
        "elapsedTime": "odrl:elapsedTime",
        "event": "odrl:event",
        "fileFormat": "odrl:fileFormat",
        "industry": "odrl:industry:",
        "language": "odrl:language",
        "media": "odrl:media",
        "meteredTime": "odrl:meteredTime",
        "payAmount": "odrl:payAmount",
        "percentage": "odrl:percentage",
        "product": "odrl:product",
        "purpose": "odrl:purpose",
        "recipient": "odrl:recipient",
        "relativePosition": "odrl:relativePosition",
        "relativeSpatialPosition": "odrl:relativeSpatialPosition",
        "relativeTemporalPosition": "odrl:relativeTemporalPosition",
        "relativeSize": "odrl:relativeSize",
        "resolution": "odrl:resolution",
        "spatial": "odrl:spatial",
        "spatialCoordinates": "odrl:spatialCoordinates",
        "systemDevice": "odrl:systemDevice",
        "timeInterval": "odrl:timeInterval",
        "unitOfCount": "odrl:unitOfCount",
        "version": "odrl:version",
        "virtualLocation": "odrl:virtualLocation",
        "eq": "odrl:eq",
        "gt": "odrl:gt",
        "gteq": "odrl:gteq",
        "lt": "odrl:lt",
        "lteq": "odrl:lteq",
        "neq": "odrl:neg",
        "isA": "odrl:isA",
        "hasPart": "odrl:hasPart",
        "isPartOf": "odrl:isPartOf",
        "isAllOf": "odrl:isAllOf",
        "isAnyOf": "odrl:isAnyOf",
        "isNoneOf": "odrl:isNoneOf",
        "or": "odrl:or",
        "xone": "odrl:xone",
        "and": "odrl:and",
        "andSequence": "odrl:andSequence",
        "policyUsage": "odrl:policyUsage",
    }
}



from unittest import IsolatedAsyncioTestCase

from ......messaging.decorators.attach_decorator import AttachDecorator
from ......tests import mock
from ......utils.testing import create_test_profile
from ...message_types import ATTACHMENT_FORMAT, CRED_20_PROPOSAL
from ...messages.cred_format import V20CredFormat
from ...messages.cred_proposal import V20CredProposal
from ...messages.inner.cred_preview import V20CredAttrSpec, V20CredPreview
from .. import cred_ex_record as test_module
from ..cred_ex_record import V20CredExRecord

TEST_DID = "LjgpST2rjsoxYegQDRm7EL"
SCHEMA_NAME = "bc-reg"
SCHEMA_TXN = 12
SCHEMA_ID = f"{TEST_DID}:2:{SCHEMA_NAME}:1.0"
SCHEMA = {
    "ver": "1.0",
    "id": SCHEMA_ID,
    "name": SCHEMA_NAME,
    "version": "1.0",
    "attrNames": ["legalName", "jurisdictionId", "incorporationDate"],
    "seqNo": SCHEMA_TXN,
}
CRED_DEF_ID = f"{TEST_DID}:3:CL:12:tag1"
CRED_PREVIEW = V20CredPreview(
    attributes=(
        V20CredAttrSpec.list_plain({"test": "123", "hello": "world"})
        + [V20CredAttrSpec(name="icon", value="cG90YXRv", mime_type="image/png")]
    )
)
INDY_FILTER = {
    "schema_id": SCHEMA_ID,
    "cred_def_id": CRED_DEF_ID,
}


class TestV20CredExRecord(IsolatedAsyncioTestCase):
    async def test_record(self):
        same = [
            V20CredExRecord(
                cred_ex_id="dummy-0",
                thread_id="thread-0",
                initiator=V20CredExRecord.INITIATOR_SELF,
                role=V20CredExRecord.ROLE_ISSUER,
            )
        ] * 2
        diff = [
            V20CredExRecord(
                cred_ex_id="dummy-1",
                initiator=V20CredExRecord.INITIATOR_SELF,
                role=V20CredExRecord.ROLE_ISSUER,
            ),
            V20CredExRecord(
                cred_ex_id="dummy-0",
                thread_id="thread-1",
                initiator=V20CredExRecord.INITIATOR_SELF,
                role=V20CredExRecord.ROLE_ISSUER,
            ),
            V20CredExRecord(
                cred_ex_id="dummy-0",
                thread_id="thread-1",
                initiator=V20CredExRecord.INITIATOR_EXTERNAL,
                role=V20CredExRecord.ROLE_ISSUER,
            ),
        ]

        for i in range(len(same) - 1):
            for j in range(i, len(same)):
                assert same[i] == same[j]

        for i in range(len(diff) - 1):
            for j in range(i, len(diff)):
                assert diff[i] == diff[j] if i == j else diff[i] != diff[j]

        assert not same[0].cred_preview  # cover non-proposal's non-preview

    def test_serde(self):
        """Test de/serialization."""

        cred_proposal = V20CredProposal(
            comment="Hello World",
            credential_preview=CRED_PREVIEW,
            formats=[
                V20CredFormat(
                    attach_id="indy",
                    format_=ATTACHMENT_FORMAT[CRED_20_PROPOSAL][
                        V20CredFormat.Format.INDY.api
                    ],
                )
            ],
            filters_attach=[AttachDecorator.data_base64(INDY_FILTER, ident="indy")],
        )
        for proposal_arg in [cred_proposal, cred_proposal.serialize()]:
            cx_rec = V20CredExRecord(
                cred_ex_id="dummy",
                connection_id="0000...",
                thread_id="dummy-thid",
                parent_thread_id="dummy-pthid",
                initiator=V20CredExRecord.INITIATOR_EXTERNAL,
                role=V20CredExRecord.ROLE_ISSUER,
                state=V20CredExRecord.STATE_PROPOSAL_RECEIVED,
                cred_proposal=proposal_arg,
                cred_offer=None,
                cred_request=None,
                cred_issue=None,
                auto_offer=False,
                auto_issue=False,
                auto_remove=True,
                error_msg=None,
                trace=False,
            )
            assert isinstance(cx_rec.cred_proposal, V20CredProposal)
            ser = cx_rec.serialize()
            deser = V20CredExRecord.deserialize(ser)
            assert isinstance(deser.cred_proposal, V20CredProposal)

    async def test_save_error_state(self):
        self.profile = await create_test_profile()
        record = V20CredExRecord(state=None)
        assert record._last_state is None
        async with self.profile.session() as session:
            await record.save_error_state(session)  # cover short circuit

            record.state = V20CredExRecord.STATE_PROPOSAL_RECEIVED
            await record.save(session)

            with (
                mock.patch.object(record, "save", mock.CoroutineMock()) as mock_save,
                mock.patch.object(
                    test_module.LOGGER, "exception", mock.MagicMock()
                ) as mock_log_exc,
            ):
                mock_save.side_effect = test_module.StorageError()
                await record.save_error_state(session, reason="test")
                mock_log_exc.assert_called_once()



"""Test Recover."""

import hashlib
from unittest import IsolatedAsyncioTestCase

import aiohttp
import base58
import indy_vdr
import pytest
from anoncreds import RevocationRegistryDefinition

from .....tests import mock
from ....models.revocation import RevList, RevRegDef, RevRegDefValue
from ..recover import (
    RevocRecoveryException,
    _check_tails_hash_for_inconsistency,
    fetch_txns,
    generate_ledger_rrrecovery_txn,
)

GENESIS = '{"reqSignature":{},"txn":{"data":{"data":{"alias":"Node1","blskey":"4N8aUNHSgjQVgkpm8nhNEfDf6txHznoYREg9kirmJrkivgL4oSEimFF6nsQ6M41QvhM2Z33nves5vfSn9n1UwNFJBYtWVnHYMATn76vLuL3zU88KyeAYcHfsih3He6UHcXDxcaecHVz6jhCYz1P2UZn2bDVruL5wXpehgBfBaLKm3Ba","blskey_pop":"RahHYiCvoNCtPTrVtP7nMC5eTYrsUA8WjXbdhNc8debh1agE9bGiJxWBXYNFbnJXoXhWFMvyqhqhRoq737YQemH5ik9oL7R4NTTCz2LEZhkgLJzB3QRQqJyBNyv7acbdHrAT8nQ9UkLbaVL9NBpnWXBTw4LEMePaSHEw66RzPNdAX1","client_ip":"172.17.0.2","client_port":9702,"node_ip":"172.17.0.2","node_port":9701,"services":["VALIDATOR"]},"dest":"Gw6pDLhcBcoQesN72qfotTgFa7cbuqZpkX3Xo6pLhPhv"},"metadata":{"from":"Th7MpTaRZVRYnPiabds81Y"},"type":"0"},"txnMetadata":{"seqNo":1,"txnId":"fea82e10e894419fe2bea7d96296a6d46f50f93f9eeda954ec461b2ed2950b62"},"ver":"1"}\n{"reqSignature":{},"txn":{"data":{"data":{"alias":"Node2","blskey":"37rAPpXVoxzKhz7d9gkUe52XuXryuLXoM6P6LbWDB7LSbG62Lsb33sfG7zqS8TK1MXwuCHj1FKNzVpsnafmqLG1vXN88rt38mNFs9TENzm4QHdBzsvCuoBnPH7rpYYDo9DZNJePaDvRvqJKByCabubJz3XXKbEeshzpz4Ma5QYpJqjk","blskey_pop":"Qr658mWZ2YC8JXGXwMDQTzuZCWF7NK9EwxphGmcBvCh6ybUuLxbG65nsX4JvD4SPNtkJ2w9ug1yLTj6fgmuDg41TgECXjLCij3RMsV8CwewBVgVN67wsA45DFWvqvLtu4rjNnE9JbdFTc1Z4WCPA3Xan44K1HoHAq9EVeaRYs8zoF5","client_ip":"172.17.0.2","client_port":9704,"node_ip":"172.17.0.2","node_port":9703,"services":["VALIDATOR"]},"dest":"8ECVSk179mjsjKRLWiQtssMLgp6EPhWXtaYyStWPSGAb"},"metadata":{"from":"EbP4aYNeTHL6q385GuVpRV"},"type":"0"},"txnMetadata":{"seqNo":2,"txnId":"1ac8aece2a18ced660fef8694b61aac3af08ba875ce3026a160acbc3a3af35fc"},"ver":"1"}\n{"reqSignature":{},"txn":{"data":{"data":{"alias":"Node3","blskey":"3WFpdbg7C5cnLYZwFZevJqhubkFALBfCBBok15GdrKMUhUjGsk3jV6QKj6MZgEubF7oqCafxNdkm7eswgA4sdKTRc82tLGzZBd6vNqU8dupzup6uYUf32KTHTPQbuUM8Yk4QFXjEf2Usu2TJcNkdgpyeUSX42u5LqdDDpNSWUK5deC5","blskey_pop":"QwDeb2CkNSx6r8QC8vGQK3GRv7Yndn84TGNijX8YXHPiagXajyfTjoR87rXUu4G4QLk2cF8NNyqWiYMus1623dELWwx57rLCFqGh7N4ZRbGDRP4fnVcaKg1BcUxQ866Ven4gw8y4N56S5HzxXNBZtLYmhGHvDtk6PFkFwCvxYrNYjh","client_ip":"172.17.0.2","client_port":9706,"node_ip":"172.17.0.2","node_port":9705,"services":["VALIDATOR"]},"dest":"DKVxG2fXXTU8yT5N7hGEbXB3dfdAnYv1JczDUHpmDxya"},"metadata":{"from":"4cU41vWW82ArfxJxHkzXPG"},"type":"0"},"txnMetadata":{"seqNo":3,"txnId":"7e9f355dffa78ed24668f0e0e369fd8c224076571c51e2ea8be5f26479edebe4"},"ver":"1"}\n{"reqSignature":{},"txn":{"data":{"data":{"alias":"Node4","blskey":"2zN3bHM1m4rLz54MJHYSwvqzPchYp8jkHswveCLAEJVcX6Mm1wHQD1SkPYMzUDTZvWvhuE6VNAkK3KxVeEmsanSmvjVkReDeBEMxeDaayjcZjFGPydyey1qxBHmTvAnBKoPydvuTAqx5f7YNNRAdeLmUi99gERUU7TD8KfAa6MpQ9bw","blskey_pop":"RPLagxaR5xdimFzwmzYnz4ZhWtYQEj8iR5ZU53T2gitPCyCHQneUn2Huc4oeLd2B2HzkGnjAff4hWTJT6C7qHYB1Mv2wU5iHHGFWkhnTX9WsEAbunJCV2qcaXScKj4tTfvdDKfLiVuU2av6hbsMztirRze7LvYBkRHV3tGwyCptsrP","client_ip":"172.17.0.2","client_port":9708,"node_ip":"172.17.0.2","node_port":9707,"services":["VALIDATOR"]},"dest":"4PS3EDQ3dW1tci1Bp6543CfuuebjFrg36kLAUcskGfaA"},"metadata":{"from":"TWwCRQRZ2ZHMJFn9TzLp7W"},"type":"0"},"txnMetadata":{"seqNo":4,"txnId":"aa5e817d7cc626170eca175822029339a444eb0ee8f0bd20d3b0b76e566fb008"},"ver":"1"}'


rev_reg_def = RevRegDef(
    tag="tag",
    cred_def_id="CsQY9MGeD3CQP4EyuVFo5m:3:CL:14951:MYCO_Biomarker",
    value=RevRegDefValue(
        max_cred_num=100,
        public_keys={
            "accum_key": {"z": "1 0BB...386"},
        },
        tails_hash="58NNWYnVxVFzAfUztwGSNBL4551XNq6nXk56pCiKJxxt",
        tails_location="http://tails-server.com",
    ),
    issuer_id="CsQY9MGeD3CQP4EyuVFo5m",
    type="CL_ACCUM",
)


@pytest.mark.anoncreds
class TestLegacyIndyRecover(IsolatedAsyncioTestCase):
    @mock.patch.object(
        indy_vdr,
        "open_pool",
        mock.CoroutineMock(
            return_value=mock.MagicMock(
                submit_request=mock.CoroutineMock(return_value={"data": {}})
            )
        ),
    )
    async def test_fetch_txns_empty_data_from_ledger(self, *_):
        with self.assertRaises(RevocRecoveryException):
            await fetch_txns(
                GENESIS,
                "4xE68b6S5VRFrKMMG1U95M:4:4xE68b6S5VRFrKMMG1U95M:3:CL:59232:default:CL_ACCUM:4ae1cc6c-f6bd-486c-8057-88f2ce74e960",
                "CsQY9MGeD3CQP4EyuVFo5m",
            )

    @mock.patch.object(
        RevocationRegistryDefinition,
        "load",
        return_value=rev_reg_def.value,
    )
    @mock.patch.object(
        indy_vdr,
        "open_pool",
        mock.CoroutineMock(
            return_value=mock.MagicMock(
                submit_request=mock.CoroutineMock(
                    return_value={
                        "data": {
                            "ver": "1.0",
                            "value": {
                                "accum_to": {},
                                "revoked": [1, 0, 1, 0],
                            },
                        }
                    }
                )
            )
        ),
    )
    @mock.patch(
        "acapy_agent.anoncreds.default.legacy_indy.recover._check_tails_hash_for_inconsistency"
    )
    async def test_fetch_txns(self, *_):
        result = await fetch_txns(
            GENESIS,
            "4xE68b6S5VRFrKMMG1U95M:4:4xE68b6S5VRFrKMMG1U95M:3:CL:59232:default:CL_ACCUM:4ae1cc6c-f6bd-486c-8057-88f2ce74e960",
            "CsQY9MGeD3CQP4EyuVFo5m",
        )
        assert isinstance(result, tuple)

    @mock.patch(
        "acapy_agent.anoncreds.default.legacy_indy.recover.fetch_txns",
        mock.CoroutineMock(
            return_value=(
                {
                    "ver": "1.0",
                    "value": {
                        "accum": "2 124C594B6B20E41B681E92B2C43FD165EA9E68BC3C9D63A82C8893124983CAE94 21 124C5341937827427B0A3A32113BD5E64FB7AB39BD3E5ABDD7970874501CA4897 6 5438CB6F442E2F807812FD9DC0C39AFF4A86B1E6766DBB5359E86A4D70401B0F 4 39D1CA5C4716FFC4FE0853C4FF7F081DFD8DF8D2C2CA79705211680AC77BF3A1 6 70504A5493F89C97C225B68310811A41AD9CD889301F238E93C95AD085E84191 4 39582252194D756D5D86D0EED02BF1B95CE12AED2FA5CD3C53260747D891993C"
                    },
                },
                {0, 1, 2},
            )
        ),
    )
    async def test_generate_ledger_rrrecovery_txn(self):
        # Has updates
        result = await generate_ledger_rrrecovery_txn(
            GENESIS,
            RevList(
                issuer_id="CsQY9MGeD3CQP4EyuVFo5m",
                current_accumulator="21 124C594B6B20E41B681E92B2C43FD165EA9E68BC3C9D63A82C8893124983CAE94 21 124C5341937827427B0A3A32113BD5E64FB7AB39BD3E5ABDD7970874501CA4897 6 5438CB6F442E2F807812FD9DC0C39AFF4A86B1E6766DBB5359E86A4D70401B0F 4 39D1CA5C4716FFC4FE0853C4FF7F081DFD8DF8D2C2CA79705211680AC77BF3A1 6 70504A5493F89C97C225B68310811A41AD9CD889301F238E93C95AD085E84191 4 39582252194D756D5D86D0EED02BF1B95CE12AED2FA5CD3C53260747D891993C",
                revocation_list=[1, 1, 1, 1],
                timestamp=1669640864487,
                rev_reg_def_id="4xE68b6S5VRFrKMMG1U95M:4:4xE68b6S5VRFrKMMG1U95M:3:CL:59232:default:CL_ACCUM:4ae1cc6c-f6bd-486c-8057-88f2ce74e960",
            ),
        )
        assert result != {}
        # Doesn't have updates
        result = await generate_ledger_rrrecovery_txn(
            GENESIS,
            RevList(
                issuer_id="CsQY9MGeD3CQP4EyuVFo5m",
                current_accumulator="21 124C594B6B20E41B681E92B2C43FD165EA9E68BC3C9D63A82C8893124983CAE94 21 124C5341937827427B0A3A32113BD5E64FB7AB39BD3E5ABDD7970874501CA4897 6 5438CB6F442E2F807812FD9DC0C39AFF4A86B1E6766DBB5359E86A4D70401B0F 4 39D1CA5C4716FFC4FE0853C4FF7F081DFD8DF8D2C2CA79705211680AC77BF3A1 6 70504A5493F89C97C225B68310811A41AD9CD889301F238E93C95AD085E84191 4 39582252194D756D5D86D0EED02BF1B95CE12AED2FA5CD3C53260747D891993C",
                revocation_list=[1, 1, 1, 0],
                timestamp=1669640864487,
                rev_reg_def_id="4xE68b6S5VRFrKMMG1U95M:4:4xE68b6S5VRFrKMMG1U95M:3:CL:59232:default:CL_ACCUM:4ae1cc6c-f6bd-486c-8057-88f2ce74e960",
            ),
        )
        assert result == {}

        # Logs waring when ledger has revoked indexes not in wallet
        with mock.patch(
            "acapy_agent.anoncreds.default.legacy_indy.recover.LOGGER"
        ) as mock_logger:
            result = await generate_ledger_rrrecovery_txn(
                GENESIS,
                RevList(
                    issuer_id="CsQY9MGeD3CQP4EyuVFo5m",
                    current_accumulator="21 124C594B6B20E41B681E92B2C43FD165EA9E68BC3C9D63A82C8893124983CAE94 21 124C5341937827427B0A3A32113BD5E64FB7AB39BD3E5ABDD7970874501CA4897 6 5438CB6F442E2F807812FD9DC0C39AFF4A86B1E6766DBB5359E86A4D70401B0F 4 39D1CA5C4716FFC4FE0853C4FF7F081DFD8DF8D2C2CA79705211680AC77BF3A1 6 70504A5493F89C97C225B68310811A41AD9CD889301F238E93C95AD085E84191 4 39582252194D756D5D86D0EED02BF1B95CE12AED2FA5CD3C53260747D891993C",
                    revocation_list=[1, 0, 0, 0],
                    timestamp=1669640864487,
                    rev_reg_def_id="4xE68b6S5VRFrKMMG1U95M:4:4xE68b6S5VRFrKMMG1U95M:3:CL:59232:default:CL_ACCUM:4ae1cc6c-f6bd-486c-8057-88f2ce74e960",
                ),
            )
            assert mock_logger.warning.called
            assert result == {}

    @mock.patch.object(
        aiohttp,
        "ClientSession",
        mock.MagicMock(
            return_value=mock.MagicMock(
                get=mock.CoroutineMock(
                    return_value=mock.MagicMock(
                        read=mock.CoroutineMock(return_value=b"some data")
                    )
                )
            )
        ),
    )
    @mock.patch.object(
        base58,
        "b58encode",
        side_effect=[
            b"58NNWYnVxVFzAfUztwGSNBL4551XNq6nXk56pCiKJxxt",
            b"58NNWYnVxVFzAfUztwGSNBL4551XNq6nXk56pCiKJxxz",
        ],
    )
    @mock.patch.object(
        hashlib,
        "sha256",
        return_value=mock.MagicMock(digest=mock.MagicMock()),
    )
    async def test_check_tails_hash_for_inconsistency(self, *_):
        # Matches
        await _check_tails_hash_for_inconsistency(
            "http://tails-server.com", "58NNWYnVxVFzAfUztwGSNBL4551XNq6nXk56pCiKJxxt"
        )
        # Mismatch
        with self.assertRaises(RevocRecoveryException):
            await _check_tails_hash_for_inconsistency(
                "http://tails-server.com",
                "58NNWYnVxVFzAfUztwGSNBL4551XNq6nXk56pCiKJxxt",
            )
